"""
Flitter controller driver API
"""

import math

from loguru import logger

from ...clock import TapTempo
from ...model import Vector, Node


class Control:
    def __init__(self, driver, control_id):
        self._initialised = False
        self.driver = driver
        self.control_id = control_id
        self.reset()

    def reset(self):
        """Set the control state back to uninitialised/default values."""
        if self._initialised:
            logger.trace("De-initialised {}({!r})", self.__class__.__name__, self.control_id)
        self._initialised = False
        self._state_prefix = None
        self._label = None
        self._color = None
        self._action = None

    def primary_state_value(self):
        raise NotImplementedError()

    def update(self, node: Node, now: float):
        """Update the configuration of the control to match the tree node."""
        changed = False
        action = node.get('action')
        if action is not None:
            action = action[0] if len(action) == 1 else tuple(action)
        if action != self._action:
            self._state_prefix = None
            self._action = action
            changed = True
        elif (state_prefix := node.get('state')) != self._state_prefix:
            self._state_prefix = state_prefix
            changed = True
        if (label := node.get('label', 1, str)) != self._label:
            self._label = label
            changed = True
        if (color := node.get('color', 3, float)) != self._color:
            self._color = color
            changed = True
        if not self._initialised:
            logger.trace("Initialised {}({!r})", self.__class__.__name__, self.control_id)
        self._initialised = True
        return changed

    def update_representation(self):
        """Update the device such that any display/LEDs/whatever match the
           current configuration and state of the control."""
        raise NotImplementedError()

    def update_state(self):
        """Update the program state dictionary and any other engine state
           (such as the beat counter tempo) to match the control state."""
        if self._state_prefix:
            self.driver.engine.state[self._state_prefix] = self.primary_state_value()


class PositionControl(Control):
    DEFAULT_LAG = 1/16

    def reset(self):
        super().reset()
        self._lag = None
        self._lower = None
        self._upper = None
        self._origin = None
        self._raw_position = None
        self._position = None
        self._position_time = None
        self._wrap = None

    @property
    def raw_divisor(self):
        raise NotImplementedError()

    @property
    def position(self):
        if self._position is None:
            return None
        if self._wrap:
            return self._lower + (self._position - self._lower) % (self._upper - self._lower)
        return min(max(self._lower, self._position), self._upper)

    def primary_state_value(self):
        return self.position

    def update(self, node, now):
        changed = super().update(node, now)
        if self._action is not None:
            match self._action:
                case 'tempo':
                    raw_position = self.driver.engine.counter.tempo * 2
                    if raw_position != self._raw_position:
                        self._raw_position = raw_position
                        changed = True
                    self._lag = None
                    self._lower = None
                    self._upper = None
                    self._origin = None
                    self._wrap = None
                    return changed
        if (lag := node.get('lag', 1, float, self.DEFAULT_LAG)) != self._lag:
            self._lag = lag
            changed = True
        if (lower := node.get('lower', 1, float, 0)) != self._lower:
            self._lower = lower
            changed = True
        if (upper := max(self._lower, node.get('upper', 1, float, 1))) != self._upper:
            self._upper = upper
            changed = True
        default_origin = 0 if self._lower < 0 < self._upper else self._lower
        if (origin := min(max(self._lower, node.get('origin', 1, float, default_origin)), self._upper)) != self._origin:
            self._origin = origin
            changed = True
        if (wrap := node.get('wrap', 1, bool, False)) != self._wrap:
            self._wrap = wrap
            changed = True
        if self._raw_position is not None:
            position_range = self._upper - self._lower
            position = self._raw_position / self.raw_divisor * position_range + self._lower
            if self._position is not None and abs(position - self._position) > position_range / 1000:
                delta = self.driver.engine.counter.beat_at_time(now) - self.driver.engine.counter.beat_at_time(self._position_time)
                alpha = math.exp(-delta / self._lag) if self._lag > 0 else 0
                self._position = self._position * alpha + position * (1 - alpha)
                self._position_time = now
                changed = True
            elif self._position != position:
                if self._wrap:
                    self._raw_position = self._raw_position % self.raw_divisor
                    self._position = self._lower + (position - self._lower) % (self._upper - self._lower)
                else:
                    self._position = position
                self._position_time = None
                changed = True
        return changed

    def handle_raw_position_change(self, raw_position, timestamp):
        if not self._initialised:
            return False
        if raw_position != self._raw_position:
            self._raw_position = raw_position
            self._position_time = timestamp
            if self._action == 'tempo' and self._raw_position != self.driver.engine.counter.tempo * 2:
                self.driver.engine.counter.tempo = self._raw_position / 2
            return True
        return False


class SettablePositionControl(PositionControl):
    def reset(self):
        super().reset()
        self._initial = None

    @property
    def initial_raw_position(self):
        if self._initialised:
            position_range = self._upper - self._lower
            position = min(max(self._lower, self._initial), self._upper)
            return (position - self._lower) / position_range * self.raw_divisor if position_range else 0
        return None

    def update(self, node, now):
        changed = super().update(node, now)
        if (initial := node.get('initial', 1, float, self._origin)) != self._initial:
            self._initial = initial
            changed = True
        if self._raw_position is None:
            if self._state_prefix and self._state_prefix in self.driver.engine.state:
                initial = float(self.driver.engine.state[self._state_prefix])
            else:
                initial = self._initial
            self._position = min(max(self._lower, initial), self._upper)
            self._position_time = now
            position_range = self._upper - self._lower
            self._raw_position = (self._position - self._lower) / position_range * self.raw_divisor if position_range else 0
            changed = True
        return changed

    def handle_raw_position_reset(self, timestamp):
        if not self._initialised or self._action is not None:
            return False
        raw_position = self.initial_raw_position
        if self._wrap:
            raw_divisor = self.raw_divisor
            delta = raw_position - self._raw_position
            while delta > raw_divisor//2:
                raw_position -= raw_divisor
                delta -= raw_divisor
            while delta < -raw_divisor//2:
                raw_position += raw_divisor
                delta += raw_divisor
        return self.handle_raw_position_change(raw_position, timestamp)


class EncoderControl(SettablePositionControl):
    STYLES = {'volume', 'pan', 'continuous'}

    def reset(self):
        super().reset()
        self._style = None
        self._turns = None

    @property
    def position(self):
        if self._wrap:
            return self._lower + (self._position - self._lower) % (self._upper - self._lower)
        elif self._style != 'continuous':
            return min(max(self._lower, self._position), self._upper)
        else:
            return self._position

    def update(self, node, now):
        changed = False
        if self._action is None and (turns := node.get('turns', 1, float, 1)) != self._turns:
            self._turns = turns
            changed = True
        if super().update(node, now):
            changed = True
        default_style = 'volume' if self._origin == self._lower else 'pan'
        style = node.get('style', 1, str, default_style).lower()
        if style not in self.STYLES:
            style = default_style
        if style != self._style:
            if self._style == 'continuous':
                raw_position = min(max(0, self._raw_position), self.raw_divisor)
                if raw_position != self._raw_position:
                    self._raw_position = raw_position
                    position_range = self._upper - self._lower
                    self._position = self._raw_position / self.raw_divisor * position_range + self._lower
                    self._position_time = now
            self._style = style
            changed = True
        return changed

    def handle_turn(self, delta, timestamp):
        if not self._initialised or delta == 0:
            return False
        raw_position = self._raw_position + delta
        if self._style != 'continuous' and not self._wrap:
            raw_position = min(max(0, raw_position), self.raw_divisor)
        return self.handle_raw_position_change(raw_position, timestamp)


class ButtonControl(Control):
    ToggleGroups = {}

    def reset(self):
        super().reset()
        self._pushed = None
        self._push_time = None
        self._release_time = None
        self._toggle = None
        self._toggled = None
        self._toggled_changed = None
        self._toggle_time = None
        self._toggle_group = None
        self._action_control = None
        self._action_control_color = None
        self._action_can_trigger = None
        self._action_triggered = None
        self._tap_tempo = None
        self._tap_time = None

    def primary_state_value(self):
        return self._pushed if not self._toggle else self._toggled

    def update(self, node, now):
        changed = super().update(node, now)
        if self._action is not None:
            triggered = self._action_can_trigger and self._action_triggered
            action_control = None
            action_can_trigger = False
            action_control_color = None
            match self._action:
                case 'next':
                    if triggered:
                        self.driver.engine.next_page()
                    return self.driver.engine.has_next_page()
                case 'previous':
                    if triggered:
                        self.driver.engine.previous_page()
                    return self.driver.engine.has_previous_page()
                case 'reset', kind, *control_id:
                    control = self.driver.get_control(kind, Vector(control_id))
                    if control is not None and control._initialised and isinstance(control, SettablePositionControl):
                        action_control = control
                        if triggered:
                            if control.handle_raw_position_reset(now):
                                control.update_representation()
                        action_control_color = control._color
                        action_can_trigger = control._raw_position != control.initial_raw_position
                case 'tap_tempo', kind, *control_id:
                    control = self.driver.get_control(kind, Vector(control_id))
                    if control is not None and control._initialised and isinstance(control, TouchControl):
                        action_control = control
                        if triggered:
                            if self._tap_tempo is None:
                                self._tap_tempo = TapTempo()
                            else:
                                self._tap_tempo.apply(self.driver.engine.counter, now)
                                self._tap_tempo = None
                                self.driver._screen_update_requested = True
                        else:
                            if control._touched_time != self._tap_time:
                                self._tap_time = control._touched_time
                                if self._tap_tempo is not None:
                                    self._tap_tempo.tap(self._tap_time)
                        action_control_color = control._color
                        action_can_trigger = True
                    else:
                        self._tap_tempo = None
                        self._tap_time = None
                case _:
                    self._tap_tempo = None
                    self._tap_time = None
            if action_control != self._action_control:
                self._action_control = action_control
                changed = True
            if action_can_trigger != self._action_can_trigger:
                self._action_can_trigger = action_can_trigger
                changed = True
            if action_control_color != self._action_control_color:
                self._action_control_color = action_control_color
                changed = True
            self._action_triggered = False
            self._toggled = None
            self._toggle_time = None
            self._toggled_changed = None
            return changed
        toggle = node.get('toggle', 1, bool, False)
        group = node.get('group')
        if group is not None:
            group = tuple(group)
        if toggle != self._toggle:
            self._toggle = toggle
            self._toggled = None
            self._toggle_time = None
            self._toggled_changed = None
            changed = True
        if self._toggle and self._toggled is None:
            key = self._state_prefix + ['toggled'] if self._state_prefix else None
            if key and key in self.driver.engine.state:
                self._toggled = bool(self.driver.engine.state[key])
                toggled_beat = float(self.driver.engine.state[key + ['beat']])
                self._toggle_time = self.driver.engine.counter.time_at_beat(toggled_beat)
            else:
                self._toggled = node.get('initial', 1, bool, False)
                self._toggle_time = now
            changed = True
        if group != self._toggle_group:
            if self._toggle_group is not None:
                buttons = self.ToggleGroups[self._toggle_group]
                buttons.remove(self)
                if not buttons:
                    del self.ToggleGroups[self._toggle_group]
            self._toggle_group = group
            self.ToggleGroups.setdefault(self._toggle_group, set()).add(self)
            changed = True
        if self._toggled_changed:
            self._toggled_changed = False
            changed = True
        return changed

    def update_state(self):
        super().update_state()
        if self._state_prefix:
            engine = self.driver.engine
            engine.state[self._state_prefix + ['pushed']] = self._pushed
            engine.state[self._state_prefix + ['released']] = not self._pushed if self._pushed is not None else None
            engine.state[self._state_prefix + ['pushed', 'beat']] = \
                engine.counter.beat_at_time(self._push_time) if self._push_time is not None else None
            engine.state[self._state_prefix + ['released', 'beat']] = \
                engine.counter.beat_at_time(self._release_time) if self._release_time is not None else None
            engine.state[self._state_prefix + ['toggled']] = self._toggled if self._toggle else None
            engine.state[self._state_prefix + ['toggled', 'beat']] = \
                engine.counter.beat_at_time(self._toggle_time) if self._toggle_time is not None else None

    def handle_push(self, pushed, timestamp):
        if not self._initialised or pushed == self._pushed:
            return False
        self._pushed = pushed
        if self._pushed:
            self._push_time = timestamp
            if self._action is not None:
                self._action_triggered = True
            elif self._toggle:
                self._toggled = not self._toggled
                self._toggle_time = timestamp
                if self._toggle_group is not None and self._toggled:
                    for button in self.ToggleGroups[self._toggle_group]:
                        if button is not self and button._toggle and button._toggled:
                            button._toggled = False
                            button._toggle_time = timestamp
                            button._toggled_changed = True
        else:
            self._release_time = timestamp
        return True


class TouchControl(Control):
    def reset(self):
        super().reset()
        self._touched = None
        self._touched_time = None
        self._untouched_time = None

    def update_state(self):
        super().update_state()
        if self._state_prefix:
            engine = self.driver.engine
            engine.state[self._state_prefix + ['touched']] = self._touched
            engine.state[self._state_prefix + ['untouched']] = not self._touched
            engine.state[self._state_prefix + ['touched', 'beat']] = \
                engine.counter.beat_at_time(self._touched_time) if self._touched_time is not None else None
            engine.state[self._state_prefix + ['untouched', 'beat']] = \
                engine.counter.beat_at_time(self._untouched_time) if self._untouched_time is not None else None

    def handle_touch(self, touched, timestamp):
        if not self._initialised or touched == self._touched:
            return False
        self._touched = touched
        if self._touched:
            self._touched_time = timestamp
        else:
            self._untouched_time = timestamp
        return True


class PressureControl(PositionControl, TouchControl, ButtonControl):
    def reset(self):
        super().reset()
        self._raw_threshold = None

    @property
    def default_raw_threshold(self):
        return self.raw_divisor // 2

    def update(self, node, now):
        if self._raw_position is None:
            self._raw_position = 0
        changed = super().update(node, now)
        if (threshold := node.get('threshold', 1, float)) is not None:
            position_range = self._upper - self._lower
            raw_threshold = (threshold - self._lower) / position_range * self.raw_divisor
        else:
            raw_threshold = self.default_raw_threshold
        if raw_threshold != self._raw_threshold:
            self._raw_threshold = raw_threshold
            changed = True
        return changed

    def handle_pressure(self, raw_position, timestamp):
        if not self._initialised or raw_position == self._raw_position:
            return False
        self._raw_position = raw_position
        changed = self.handle_push(self._raw_position >= self._raw_threshold, timestamp)
        changed = self.handle_touch(self._raw_position > 0, timestamp) or changed
        return changed


class ControllerDriver:
    def __init__(self, engine):
        self.engine = engine

    def get_default_config(self) -> list[Node]:
        """Return a list of nodes that specify a default configuration."""
        raise NotImplementedError()

    async def start(self):
        """Start the driver."""
        raise NotImplementedError()

    async def stop(self):
        """Stop the driver (or schedule it to stop)."""
        raise NotImplementedError()

    async def start_update(self, node: Node):
        """Start an update pass."""
        pass

    def get_control(self, kind: str, control_id: Vector) -> Control:
        """Return a control matching the node `kind` and `id` attribute."""
        raise NotImplementedError()

    def handle_node(self, node: Node) -> bool:
        """A second chance to handle nodes that don't appear to be controls.
           Return `False` if not handled, `True` otherwise."""
        return False

    async def finish_update(self):
        """Finish an update pass."""
        pass



"""
Behringer X-Touch Mini controller driver
"""

import asyncio

from loguru import logger

from . import driver, midi
from ...model import Vector, Node


def get_driver_class():
    return XTouchMiniDriver


DEFAULT_CONFIG = [
    Node('button', attributes={'id': Vector('a'), 'action': Vector('next')}),
    Node('button', attributes={'id': Vector('b'), 'action': Vector('previous')}),
    Node('button', attributes={'id': Vector(('rotary', 1)), 'action': Vector(('reset', 'rotary', 1))}),
    Node('button', attributes={'id': Vector(('rotary', 2)), 'action': Vector(('reset', 'rotary', 2))}),
    Node('button', attributes={'id': Vector(('rotary', 3)), 'action': Vector(('reset', 'rotary', 3))}),
    Node('button', attributes={'id': Vector(('rotary', 4)), 'action': Vector(('reset', 'rotary', 4))}),
    Node('button', attributes={'id': Vector(('rotary', 5)), 'action': Vector(('reset', 'rotary', 5))}),
    Node('button', attributes={'id': Vector(('rotary', 6)), 'action': Vector(('reset', 'rotary', 6))}),
    Node('button', attributes={'id': Vector(('rotary', 7)), 'action': Vector(('reset', 'rotary', 7))}),
    Node('button', attributes={'id': Vector(('rotary', 8)), 'action': Vector(('reset', 'rotary', 8))}),
]

NOTE_BUTTON_MAPPING = {
    32: Vector(('rotary', 1)), 33: Vector(('rotary', 2)), 34: Vector(('rotary', 3)), 35: Vector(('rotary', 4)),
    36: Vector(('rotary', 5)), 37: Vector(('rotary', 6)), 38: Vector(('rotary', 7)), 39: Vector(('rotary', 8)),
    89: Vector(1), 90: Vector(2), 40: Vector(3), 41: Vector(4), 42: Vector(5), 43: Vector(6), 44: Vector(7), 45: Vector(8),
    87: Vector(9), 88: Vector(10), 91: Vector(11), 92: Vector(12), 86: Vector(13), 93: Vector(14), 94: Vector(15), 95: Vector(16),
    84: Vector('a'), 85: Vector('b'),
}

BUTTON_LIGHT_MAPPING = {
    Vector(1): 89, Vector(2): 90, Vector(3): 40, Vector(4): 41, Vector(5): 42, Vector(6): 43, Vector(7): 44, Vector(8): 45,
    Vector(9): 87, Vector(10): 88, Vector(11): 91, Vector(12): 92, Vector(13): 86, Vector(14): 93, Vector(15): 94, Vector(16): 95,
    Vector("a"): 84, Vector("b"): 85,
}

ROTARY_CONTROLS_MAPPING = {
    Vector(1): (16, 48), Vector(2): (17, 49), Vector(3): (18, 50), Vector(4): (19, 51),
    Vector(5): (20, 52), Vector(6): (21, 53), Vector(7): (22, 54), Vector(8): (23, 55),
}

TURN_CONTROL_ROTARY_MAPPING = {turn_control: rotary_id for rotary_id, (turn_control, light_control) in ROTARY_CONTROLS_MAPPING.items()}

ALIASES = {
    ('button', Vector('mc')): Vector(9),
    ('button', Vector('reverse')): Vector(11),
    ('button', Vector('forward')): Vector(12),
    ('button', Vector('loop')): Vector(13),
    ('button', Vector('stop')): Vector(14),
    ('button', Vector('play')): Vector(15),
    ('button', Vector('record')): Vector(16),
}

SPECIAL_ACTIONS = {
    Vector("a"): 'next', Vector("b"): 'previous',
}


class XTouchMiniRotary(driver.EncoderControl):
    DEFAULT_LAG = 1/4

    def __init__(self, driver, control_id, light_control):
        super().__init__(driver, control_id)
        self._light_control = light_control

    @property
    def raw_divisor(self):
        return int(round(24 * self._turns))

    def update_representation(self):
        if self.driver._midi_port is None:
            return
        if self._initialised:
            divisor = self.raw_divisor
            if self._upper == self._lower:
                value = 0x3f
            elif self._style == 'volume':
                value = int(round(self._raw_position / divisor * 10)) + 0x21
            elif self._style == 'pan':
                value = int(round(self._raw_position / divisor * 10)) + 0x11
            else:
                value = int(self._raw_position / divisor % 1 * 11) + 0x01
            self.driver._midi_port.send_control_change(self._light_control, value)
        else:
            self.driver._midi_port.send_control_change(self._light_control, 0)


class XTouchMiniButton(driver.ButtonControl):
    def __init__(self, driver, control_id, light_note):
        super().__init__(driver, control_id)
        self._light_note = light_note

    def update_representation(self):
        if self._light_note is None or self.driver._midi_port is None:
            return
        if self._initialised:
            if self._action:
                velocity = 127 if self._action_can_trigger else 0
            elif self._color is not None:
                velocity = 127 if self._color != [0, 0, 0] else 0
            elif self._toggle:
                velocity = 127 if self._toggled else 0
            else:
                velocity = 127 if self._pushed else 0
            self.driver._midi_port.send_note_on(self._light_note, velocity)
        else:
            self.driver._midi_port.send_note_on(self._light_note, 0)


class XTouchMiniFader(driver.PositionControl):
    def reset(self):
        raw_position = self._raw_position if hasattr(self, '_raw_position') else None
        super().reset()
        self._raw_position = raw_position

    def update_representation(self):
        pass

    @property
    def raw_divisor(self):
        return 16256


class XTouchMiniDriver(driver.ControllerDriver):
    VENDOR_ID = 0x1397
    PRODUCT_ID = 0x00b3
    PORT_NAME = 'X-TOUCH MINI'

    def __init__(self, engine):
        super().__init__(engine)
        self._rotaries = {}
        self._buttons = {}
        self._sliders = {}
        self._toggle_groups = {}
        self._run_task = None
        self._midi_port = None
        for rotary_id, (_, light_control) in ROTARY_CONTROLS_MAPPING.items():
            rotary = XTouchMiniRotary(self, rotary_id, light_control)
            self._rotaries[rotary_id] = rotary
        for note, button_id in NOTE_BUTTON_MAPPING.items():
            light_note = BUTTON_LIGHT_MAPPING.get(button_id)
            button = XTouchMiniButton(self, button_id, light_note)
            self._buttons[button_id] = button
        self._sliders[Vector('main')] = XTouchMiniFader(self, Vector('main'))

    async def start(self):
        self._run_task = asyncio.create_task(self.run())

    async def stop(self):
        self._run_task.cancel()
        await self._run_task
        self._run_task = None

    def get_default_config(self):
        return DEFAULT_CONFIG

    async def run(self):
        try:
            while True:
                while not self._try_connect():
                    await asyncio.sleep(1)
                self._midi_port.send_control_change(127, 1, channel=11)  # Switch to Mackie Control mode
                self.refresh()
                logger.debug("X-Touch Mini controller driver ready")
                while True:
                    event = await self._midi_port.wait_event(1)
                    if event is None:
                        self.refresh()
                        continue
                    match event:
                        case midi.NoteOnEvent(note=note, channel=0) if note in NOTE_BUTTON_MAPPING:
                            button = self._buttons[NOTE_BUTTON_MAPPING[note]]
                            if button.handle_push(event.velocity == 127, event.timestamp):
                                button.update_representation()
                        case midi.ControlChangeEvent(control=control, value=value, channel=0) if control in TURN_CONTROL_ROTARY_MAPPING:
                            rotary = self._rotaries[TURN_CONTROL_ROTARY_MAPPING[control]]
                            if rotary.handle_turn(64 - value if value > 64 else value, event.timestamp):
                                rotary.update_representation()
                        case midi.PitchBendEvent(channel=8):
                            self._sliders[Vector('main')].handle_raw_position_change(event.value, event.timestamp)
                        case _:
                            logger.warning("Unhandled MIDI event: {}", event)
        except asyncio.CancelledError:
            if self._midi_port is not None:
                self._midi_port.send_control_change(127, 0, channel=11)
                self._midi_port.close()
                self._midi_port = None
        except Exception as exc:
            logger.opt(exception=exc).error("Unhandled exception in X-Touch mini controller main loop")

    def refresh(self):
        for rotary in self._rotaries.values():
            rotary.update_representation()
        for button in self._buttons.values():
            button.update_representation()
        for slider in self._sliders.values():
            slider.update_representation()

    def _try_connect(self):
        try:
            self._midi_port = midi.MidiPort(self.PORT_NAME)
        except ValueError:
            return False
        else:
            return True

    def get_control(self, kind, control_id):
        control_id = ALIASES.get((kind, control_id), control_id)
        if kind == 'rotary':
            return self._rotaries.get(control_id)
        if kind == 'button':
            return self._buttons.get(control_id)
        if kind == 'slider':
            return self._sliders.get(control_id)



"""
Tests of the model.Matrix44 class
"""

import math
import unittest

from flitter.model import Matrix44, Matrix33


def all_isclose(xs, ys, rel_tol=1e-9, abs_tol=0):
    for x, y in zip(xs, ys):
        if not math.isclose(x, y, rel_tol=rel_tol, abs_tol=abs_tol):
            return False
    return True


class TestMatrix44(unittest.TestCase):
    """
    Tests of the Matrix44 class

    Note that matrices are column-major as per OpenGL.
    """

    def test_construct(self):
        self.assertEqual(Matrix44(), [1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1])
        self.assertEqual(Matrix44(0), [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0])
        self.assertEqual(Matrix44(1), [1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1])
        self.assertEqual(Matrix44(2), [2, 0, 0, 0, 0, 2, 0, 0, 0, 0, 2, 0, 0, 0, 0, 2])
        self.assertEqual(Matrix44(range(16)), range(16))
        self.assertRaises(ValueError, Matrix44, "Hello world!")
        self.assertRaises(ValueError, Matrix44, [1, 2, 3])

    def test_project(self):
        self.assertTrue(all_isclose(Matrix44.project(1, 1, 1, 2), [1, 0, 0, 0, 0, 1, 0, 0, 0, 0, -3, -1, 0, 0, -4, 0]))
        self.assertTrue(all_isclose(Matrix44.project(1, 0.5, 1, 2), [1, 0, 0, 0, 0, 2, 0, 0, 0, 0, -3, -1, 0, 0, -4, 0]))
        self.assertTrue(all_isclose(Matrix44.project(1, 0.5, 1, 3), [1, 0, 0, 0, 0, 2, 0, 0, 0, 0, -2, -1, 0, 0, -3, 0]))
        self.assertTrue(all_isclose(Matrix44.project(0.5773502691896256, 0.2886751345948128, 1, 3),
                                    [1.7320508075688776, 0, 0, 0, 0, 3.4641016151377553, 0, 0, 0, 0, -2, -1, 0, 0, -3, 0]))

    def test_ortho(self):
        self.assertTrue(all_isclose(Matrix44.ortho(1, 2, 0, 2), [1, 0, 0, 0, 0, 1, 0, 0, 0, 0, -1, 0, 0, 0, -1, 1]))
        self.assertTrue(all_isclose(Matrix44.ortho(2, 2, 0, 2), [1, 0, 0, 0, 0, 2, 0, 0, 0, 0, -1, 0, 0, 0, -1, 1]))
        self.assertTrue(all_isclose(Matrix44.ortho(2, 1, 0, 2), [2, 0, 0, 0, 0, 4, 0, 0, 0, 0, -1, 0, 0, 0, -1, 1]))
        self.assertTrue(all_isclose(Matrix44.ortho(2, 1, 1, 2), [2, 0, 0, 0, 0, 4, 0, 0, 0, 0, -2, 0, 0, 0, -3, 1]))
        self.assertTrue(all_isclose(Matrix44.ortho(2, 1, 1, 3), [2, 0, 0, 0, 0, 4, 0, 0, 0, 0, -1, 0, 0, 0, -2, 1]))

    def test_look(self):
        self.assertIsNone(Matrix44.look([0], None, "Hello world!"))
        self.assertTrue(all_isclose(Matrix44.look([0, 0, 0], [0, 0, -1], [0, 1, 0]), Matrix44()))
        self.assertTrue(all_isclose(Matrix44.look([0, 0, 0], [0, 0, -100], [0, 1, 0]), Matrix44()))
        self.assertTrue(all_isclose(Matrix44.look([0, 0, 0], [0, 0, -100], [1, 0, 0]),
                                    [0, 1, 0, 0, -1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1]))
        self.assertTrue(all_isclose(Matrix44.look([0, 0, 0], [0, 0, 100], [1, 0, 0]),
                                    [0, 1, 0, 0, 1, 0, 0, 0, 0, 0, -1, 0, 0, 0, 0, 1]))
        self.assertTrue(all_isclose(Matrix44.look([1, 2, 3], [1, 2, 2], [0, 1, 0]),
                                    [1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, -1, -2, -3, 1]))

    def test_translate(self):
        self.assertIsNone(Matrix44.translate(None))
        self.assertIsNone(Matrix44.translate([0, 1]))
        self.assertIsNone(Matrix44.translate("Hello world!"))
        self.assertEqual(Matrix44.translate([2]), [1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 2, 2, 2, 1])
        self.assertEqual(Matrix44.translate([1, 2, 3]), [1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 1, 2, 3, 1])

    def test_scale(self):
        self.assertIsNone(Matrix44.scale(None))
        self.assertIsNone(Matrix44.scale([0, 1]))
        self.assertIsNone(Matrix44.scale("Hello world!"))
        self.assertEqual(Matrix44.scale(2), [2, 0, 0, 0, 0, 2, 0, 0, 0, 0, 2, 0, 0, 0, 0, 1])
        self.assertEqual(Matrix44.scale([1, 2, 3]), [1, 0, 0, 0, 0, 2, 0, 0, 0, 0, 3, 0, 0, 0, 0, 1])

    def test_rotate_x(self):
        self.assertIsNone(Matrix44.rotate_x(math.nan))
        self.assertEqual(Matrix44.rotate_x(0), Matrix44())
        self.assertTrue(all_isclose(Matrix44.rotate_x(1), Matrix44(), abs_tol=1e-12))
        self.assertTrue(all_isclose(Matrix44.rotate_x(0.5), [1, 0, 0, 0, 0, -1, 0, 0, 0, 0, -1, 0, 0, 0, 0, 1], abs_tol=1e-12))

    def test_rotate_y(self):
        self.assertIsNone(Matrix44.rotate_y(math.nan))
        self.assertEqual(Matrix44.rotate_y(0), Matrix44())
        self.assertTrue(all_isclose(Matrix44.rotate_y(1), Matrix44(), abs_tol=1e-12))
        self.assertTrue(all_isclose(Matrix44.rotate_y(0.5), [-1, 0, 0, 0, 0, 1, 0, 0, 0, 0, -1, 0, 0, 0, 0, 1], abs_tol=1e-12))

    def test_rotate_z(self):
        self.assertIsNone(Matrix44.rotate_z(math.nan))
        self.assertEqual(Matrix44.rotate_z(0), Matrix44())
        self.assertTrue(all_isclose(Matrix44.rotate_z(1), Matrix44(), abs_tol=1e-12))
        self.assertTrue(all_isclose(Matrix44.rotate_z(0.5), [-1, 0, 0, 0, 0, -1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1], abs_tol=1e-12))

    def test_rotate(self):
        self.assertEqual(Matrix44.rotate(None), Matrix44())
        self.assertEqual(Matrix44.rotate("Hello world!"), Matrix44())
        self.assertEqual(Matrix44.rotate([0.1, 0.2]), Matrix44())
        self.assertEqual(Matrix44.rotate(0), Matrix44())
        self.assertTrue(all_isclose(Matrix44.rotate(1), Matrix44(), abs_tol=1e-12))
        self.assertTrue(all_isclose(Matrix44.rotate(0.5), Matrix44(), abs_tol=1e-12))
        self.assertTrue(all_isclose(Matrix44.rotate([0.5, 0.5, 0.5]), Matrix44(), abs_tol=1e-12))
        self.assertTrue(all_isclose(Matrix44.rotate([0.5, 0, 0]), [1, 0, 0, 0, 0, -1, 0, 0, 0, 0, -1, 0, 0, 0, 0, 1], abs_tol=1e-12))
        self.assertTrue(all_isclose(Matrix44.rotate([0, 0.5, 0]), [-1, 0, 0, 0, 0, 1, 0, 0, 0, 0, -1, 0, 0, 0, 0, 1], abs_tol=1e-12))
        self.assertTrue(all_isclose(Matrix44.rotate([0, 0, 0.5]), [-1, 0, 0, 0, 0, -1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1], abs_tol=1e-12))

    def test_shear_x(self):
        self.assertIsNone(Matrix44.shear_x(None))
        self.assertIsNone(Matrix44.shear_x([0, 1, 2]))
        self.assertEqual(Matrix44.shear_x(0), Matrix44())
        self.assertEqual(Matrix44.shear_x([0, 0]), Matrix44())
        self.assertEqual(Matrix44.shear_x([1, 0]), [1, 0, 0, 0,  1, 1, 0, 0,  0, 0, 1, 0,  0, 0, 0, 1])
        self.assertEqual(Matrix44.shear_x([0, 1]), [1, 0, 0, 0,  0, 1, 0, 0,  1, 0, 1, 0,  0, 0, 0, 1])
        self.assertEqual(Matrix44.shear_x([1, 1]), [1, 0, 0, 0,  1, 1, 0, 0,  1, 0, 1, 0,  0, 0, 0, 1])
        self.assertEqual(Matrix44.shear_x(1),      [1, 0, 0, 0,  1, 1, 0, 0,  1, 0, 1, 0,  0, 0, 0, 1])
        self.assertEqual(Matrix44.shear_x([0.5, 0.75]), [1, 0, 0, 0,  0.5, 1, 0, 0,  0.75, 0, 1, 0,  0, 0, 0, 1])

    def test_shear_y(self):
        self.assertIsNone(Matrix44.shear_y(None))
        self.assertIsNone(Matrix44.shear_y([0, 1, 2]))
        self.assertEqual(Matrix44.shear_y(0), Matrix44())
        self.assertEqual(Matrix44.shear_y([0, 0]), Matrix44())
        self.assertEqual(Matrix44.shear_y([1, 0]), [1, 1, 0, 0,  0, 1, 0, 0,  0, 0, 1, 0,  0, 0, 0, 1])
        self.assertEqual(Matrix44.shear_y([0, 1]), [1, 0, 0, 0,  0, 1, 0, 0,  0, 1, 1, 0,  0, 0, 0, 1])
        self.assertEqual(Matrix44.shear_y([1, 1]), [1, 1, 0, 0,  0, 1, 0, 0,  0, 1, 1, 0,  0, 0, 0, 1])
        self.assertEqual(Matrix44.shear_y(1),      [1, 1, 0, 0,  0, 1, 0, 0,  0, 1, 1, 0,  0, 0, 0, 1])
        self.assertEqual(Matrix44.shear_y([0.5, 0.75]), [1, 0.5, 0, 0,  0, 1, 0, 0,  0, 0.75, 1, 0,  0, 0, 0, 1])

    def test_shear_z(self):
        self.assertIsNone(Matrix44.shear_z(None))
        self.assertIsNone(Matrix44.shear_z([0, 1, 2]))
        self.assertEqual(Matrix44.shear_z(0), Matrix44())
        self.assertEqual(Matrix44.shear_z([0, 0]), Matrix44())
        self.assertEqual(Matrix44.shear_z([1, 0]), [1, 0, 1, 0,  0, 1, 0, 0,  0, 0, 1, 0,  0, 0, 0, 1])
        self.assertEqual(Matrix44.shear_z([0, 1]), [1, 0, 0, 0,  0, 1, 1, 0,  0, 0, 1, 0,  0, 0, 0, 1])
        self.assertEqual(Matrix44.shear_z([1, 1]), [1, 0, 1, 0,  0, 1, 1, 0,  0, 0, 1, 0,  0, 0, 0, 1])
        self.assertEqual(Matrix44.shear_z(1),      [1, 0, 1, 0,  0, 1, 1, 0,  0, 0, 1, 0,  0, 0, 0, 1])
        self.assertEqual(Matrix44.shear_z([0.5, 0.75]), [1, 0, 0.5, 0,  0, 1, 0.75, 0,  0, 0, 1, 0,  0, 0, 0, 1])

    def test_mmul(self):
        self.assertEqual(Matrix44() @ Matrix44(), Matrix44())
        self.assertEqual(Matrix44.translate([1, 2, 3]) @ Matrix44(), Matrix44.translate([1, 2, 3]))
        self.assertEqual(Matrix44() @ Matrix44.translate([1, 2, 3]), Matrix44.translate([1, 2, 3]))
        self.assertEqual(Matrix44.translate([1, 2, 3]) @ Matrix44.translate([1, 2, 3]), Matrix44.translate([2, 4, 6]))
        self.assertEqual(Matrix44.translate([1, 2, 3]) @ Matrix44.scale(2), [2, 0, 0, 0, 0, 2, 0, 0, 0, 0, 2, 0, 1, 2, 3, 1])
        self.assertEqual(Matrix44.scale(2) @ Matrix44.translate([1, 2, 3]), [2, 0, 0, 0, 0, 2, 0, 0, 0, 0, 2, 0, 2, 4, 6, 1])

    def test_vmul(self):
        self.assertEqual(Matrix44() @ None, None)
        self.assertEqual(Matrix44() @ [], None)
        self.assertEqual(Matrix44() @ [1], None)
        self.assertEqual(Matrix44() @ [1, 2], None)
        self.assertEqual(Matrix44() @ [1, 2, 3], [1, 2, 3])
        self.assertEqual(Matrix44() @ [1, 2, 3, 1], [1, 2, 3, 1])
        self.assertEqual(Matrix44.translate([1, 2, 3]) @ [0, 0, 0], [1, 2, 3])
        self.assertEqual(Matrix44.translate([1, 2, 3]) @ [-1, -2, -3, 1], [0, 0, 0, 1])
        self.assertEqual(Matrix44.scale(2) @ [-1, -2, -3, 1], [-2, -4, -6, 1])

    def test_inverse(self):
        a = Matrix44.project(1, 1, 1, 3)
        b = Matrix44.scale([5, 6, -3])
        c = Matrix44.look([1, 2, 3], [1, 2, 2], [-1, 0, 0])
        self.assertTrue(all_isclose(a.inverse() @ a, Matrix44()))
        self.assertTrue(all_isclose(b.inverse() @ b, Matrix44()))
        self.assertTrue(all_isclose(c.inverse() @ c, Matrix44()))
        self.assertTrue(all_isclose(b.inverse() @ (a.inverse() @ (a @ (b @ c))), c))
        self.assertTrue(all_isclose((((b.inverse() @ a.inverse()) @ a) @ b) @ c, c))

    def test_transpose(self):
        self.assertEqual(Matrix44().transpose(), Matrix44())
        self.assertEqual(Matrix44(range(16)).transpose(), [0, 4, 8, 12, 1, 5, 9, 13, 2, 6, 10, 14, 3, 7, 11, 15])

    def test_matrix33(self):
        self.assertEqual(Matrix44().matrix33(), Matrix33())
        self.assertEqual(Matrix44(range(16)).matrix33(), [0, 1, 2, 4, 5, 6, 8, 9, 10])

    def test_inverse_transpose_matrix33(self):
        m = Matrix44.project(2, 1 / 6, 1, 3) @ \
            Matrix44.look([1, 2, 3], [1, 2, 2], [-1, 0, 0]) @ \
            Matrix44.scale([5, 6, -3])
        self.assertTrue(all_isclose(m.inverse_transpose_matrix33(), m.inverse().transpose().matrix33()))

    def test_repr(self):
        self.assertEqual(repr(Matrix44()), """|   1.000   0.000   0.000   0.000 |
|   0.000   1.000   0.000   0.000 |
|   0.000   0.000   1.000   0.000 |
|   0.000   0.000   0.000   1.000 |""")



"""
Flitter functional testing
"""

import asyncio
from pathlib import Path
import tempfile
import unittest

import PIL.Image
import PIL.ImageChops
import PIL.ImageStat

from flitter.engine.control import EngineController


def image_diff(ref, img):
    """Return how dissimilar two images are as a score between 0 (identical) and 1 (completely different)"""
    return PIL.ImageStat.Stat(PIL.ImageChops.difference(ref, img).convert('L')).sum[0] / (img.width * img.height * 255)


TEST_IMAGES_DIR = Path(__file__).parent.parent / 'build/test_images'
TEST_IMAGES_DIR.mkdir(mode=0o775, parents=True, exist_ok=True)

for path in TEST_IMAGES_DIR.iterdir():
    if path.is_file():
        path.unlink(missing_ok=True)


class TestRendering(unittest.TestCase):
    """
    Some simple rendering functional tests

    Scripts will be executed for a single frame and a `SIZE` image should be
    saved to the filename `OUTPUT`.
    """

    SIZE = (200, 100)

    def setUp(self):
        self.script_path = Path(tempfile.mktemp('.fl'))
        self.input_path = Path(tempfile.mktemp('.png'))
        self.output_path = Path(tempfile.mktemp('.png'))
        self.controller = EngineController(realtime=False, target_fps=1, run_time=1, offscreen=True,
                                           defined_names={'SIZE': self.SIZE,
                                                          'INPUT': str(self.input_path),
                                                          'OUTPUT': str(self.output_path)})

    def tearDown(self):
        if self.script_path.exists():
            self.script_path.unlink()
        if self.input_path.exists():
            self.input_path.unlink()
        if self.output_path.exists():
            self.output_path.unlink()

    def test_simple_canvas(self):
        """Render a simple green rectangle filling the viewport"""
        self.script_path.write_text("""
!window size=SIZE
    !record filename=OUTPUT
        !canvas
            !path
                !rect size=SIZE
                !fill color=0;1;0
""", encoding='utf8', newline='\n')
        self.controller.load_page(self.script_path)
        asyncio.run(self.controller.run())
        img = PIL.Image.open(self.output_path)
        self.assertEqual(img.size, self.SIZE)
        self.assertEqual(img.tobytes(), b'\x00\xff\x00' * self.SIZE[0] * self.SIZE[1])

    def test_simple_canvas3d(self):
        """Render a simple green emissive box filling the viewport"""
        self.script_path.write_text("""
!window size=SIZE
    !record filename=OUTPUT
        !canvas3d orthographic=true
            !box size=SIZE*(1;1;0) emissive=0;1;0
""", encoding='utf8', newline='\n')
        self.controller.load_page(self.script_path)
        asyncio.run(self.controller.run())
        img = PIL.Image.open(self.output_path)
        self.assertEqual(img.size, self.SIZE)
        self.assertEqual(img.tobytes(), b'\x00\xff\x00' * self.SIZE[0] * self.SIZE[1])

    def test_simple_shader(self):
        """Render a simple shader that returns green for all fragments"""
        self.script_path.write_text("""
!window size=SIZE
    !record filename=OUTPUT
        !shader fragment='''#version 330
                            in vec2 coord;
                            out vec4 color;
                            void main() {
                                color = vec4(0, 1, 0, 1);
                            }'''
""", encoding='utf8', newline='\n')
        self.controller.load_page(self.script_path)
        asyncio.run(self.controller.run())
        img = PIL.Image.open(self.output_path)
        self.assertEqual(img.size, self.SIZE)
        self.assertEqual(img.tobytes(), b'\x00\xff\x00' * self.SIZE[0] * self.SIZE[1])

    def test_simple_image(self):
        """Render an image node from a simple temporary image"""
        PIL.Image.new('RGB', self.SIZE, (0, 255, 0)).save(self.input_path, 'PNG')
        self.script_path.write_text("""
!window size=SIZE
    !record filename=OUTPUT
        !image filename=INPUT
""", encoding='utf8', newline='\n')
        self.controller.load_page(self.script_path)
        asyncio.run(self.controller.run())
        img = PIL.Image.open(self.output_path)
        self.assertEqual(img.size, self.SIZE)
        self.assertEqual(img.tobytes(), b'\x00\xff\x00' * self.SIZE[0] * self.SIZE[1])


class TestDocumentationDiagrams(unittest.TestCase):
    """
    Recreate the documentation diagrams and check them against the pre-calculated ones.
    Some amount of difference is expected here because the GitHub workflow tests run on
    Ubuntu and the font availability is different. There also seems to be an issue with
    Skia not doing anti-aliasing in Xvfb (Mesa).
    """

    def test_diagrams(self):
        scripts_dir = Path(__file__).parent.parent / 'docs/diagrams'
        scripts = sorted(path for path in scripts_dir.iterdir() if path.suffix == '.fl')
        self.assertTrue(len(scripts) > 0)
        for i, script in enumerate(scripts):
            with self.subTest(script=script):
                output_path = TEST_IMAGES_DIR / script.with_suffix('.png').name
                reference = PIL.Image.open(script.with_suffix('.png'))
                controller = EngineController(realtime=False, target_fps=1, run_time=1, offscreen=True,
                                              defined_names={'OUTPUT': str(output_path)})
                controller.load_page(script)
                asyncio.run(controller.run())
                output = PIL.Image.open(output_path)
                self.assertEqual(reference.size, output.size)
                self.assertLess(image_diff(reference, output), 0.002)


class TestDocumentationTutorial(unittest.TestCase):
    """
    Recreate the tutorial images and check them against the pre-calculated ones.
    """

    def test_tutorial(self):
        scripts_dir = Path(__file__).parent.parent / 'docs/tutorial_images'
        scripts = sorted(path for path in scripts_dir.iterdir() if path.suffix == '.fl')
        self.assertTrue(len(scripts) > 0)
        for i, script in enumerate(scripts):
            with self.subTest(script=script):
                output_path = TEST_IMAGES_DIR / script.with_suffix('.png').name
                reference = PIL.Image.open(script.with_suffix('.png'))
                controller = EngineController(realtime=False, target_fps=1, run_time=1, offscreen=True,
                                              defined_names={'OUTPUT': str(output_path)})
                controller.load_page(script)
                asyncio.run(controller.run())
                output = PIL.Image.open(output_path)
                self.assertEqual(reference.size, output.size)
                self.assertLess(image_diff(reference, output), 0.005)


class TestExamples(unittest.TestCase):
    """
    Recreate the examples and check them against the pre-calculated ones.
    """

    SHORT = ['bauble.fl', 'canvas3d.fl', 'solidgeometry.fl', 'textures.fl', 'video.fl']
    LONG = ['bounce.fl', 'hoops.fl', 'physics.fl', 'smoke.fl']

    def test_short_examples(self):
        scripts_dir = Path(__file__).parent.parent / 'examples'
        scripts = [scripts_dir / name for name in self.SHORT]
        for i, script in enumerate(scripts):
            with self.subTest(script=script):
                output_path = TEST_IMAGES_DIR / script.with_suffix('.png').name
                reference = PIL.Image.open(script.with_suffix('.png'))
                controller = EngineController(realtime=False, target_fps=1, run_time=1, offscreen=True,
                                              defined_names={'OUTPUT': str(output_path)})
                controller.load_page(script)
                asyncio.run(controller.run())
                output = PIL.Image.open(output_path)
                self.assertEqual(reference.size, output.size)
                self.assertLess(image_diff(reference, output), 0.001)

    def test_long_examples(self):
        scripts_dir = Path(__file__).parent.parent / 'examples'
        scripts = [scripts_dir / name for name in self.LONG]
        for i, script in enumerate(scripts):
            with self.subTest(script=script):
                output_path = TEST_IMAGES_DIR / script.with_suffix('.png').name
                reference = PIL.Image.open(script.with_suffix('.png'))
                controller = EngineController(realtime=False, target_fps=10, run_time=1, offscreen=True,
                                              defined_names={'OUTPUT': str(output_path)})
                controller.load_page(script)
                asyncio.run(controller.run())
                output = PIL.Image.open(output_path)
                self.assertEqual(reference.size, output.size)
                self.assertLess(image_diff(reference, output), 0.004)



"""
Flitter generic controller framework
"""

import asyncio
import importlib

from loguru import logger

from . import midi


class Controller:
    VIRTUAL_MIDI_PORT = None
    DRIVER_CACHE = {}

    @classmethod
    def open_virtual_midi_port(cls):
        """Hold a virtual MIDI port open to stop RTMIDI on OS X releasing the
           main context when ports are closed and throwing subsequent errors
           when attempting to open a new port."""
        if cls.VIRTUAL_MIDI_PORT is None:
            cls.VIRTUAL_MIDI_PORT = midi.MidiPort('flitter', virtual=True)

    def __init__(self, **kwargs):
        logger.debug("Create controller")
        self.open_virtual_midi_port()
        self.driver = None
        self.controls = {}
        self.unknown = set()

    def purge(self):
        while self.controls:
            key, control = self.controls.popitem()
            control.reset()
            control.update_representation()

    def destroy(self):
        if self.driver is not None:
            self.purge()
            asyncio.create_task(self.driver.stop())
            self.driver = None

    async def update(self, engine, node, clock, **kwargs):
        driver = node.get('driver', 1, str)
        if driver in self.DRIVER_CACHE:
            driver_class = self.DRIVER_CACHE[driver]
        elif driver:
            try:
                driver_module = importlib.import_module(f'.{driver}', __package__)
                driver_class = driver_module.get_driver_class()
            except (ImportError, NameError):
                logger.warning("Unable to import controller driver: {}", driver)
                driver_class = None
            self.DRIVER_CACHE[driver] = driver_class
        else:
            driver_class = None
        if self.driver is not None and (driver_class is None or not isinstance(self.driver, driver_class)):
            await self.driver.stop()
            self.driver = None
        if self.driver is None:
            if driver_class is not None:
                self.driver = driver_class(engine)
                await self.driver.start()
            else:
                return
        await self.driver.start_update(node)
        controls = {}
        unknown = set()
        for child in list(node.children) + self.driver.get_default_config():
            if 'id' in child:
                control_id = child['id']
                control = self.driver.get_control(child.kind, control_id)
                if control is not None:
                    key = child.kind, control_id
                    if key in controls:
                        continue
                    controls[key] = control
                    if key in self.controls:
                        del self.controls[key]
                    if control.update(child, clock):
                        control.update_representation()
                    control.update_state()
                    continue
            if not self.driver.handle_node(child):
                unknown.add(child.kind)
        for kind in unknown.difference(self.unknown):
            logger.warning("Unexpected '{}' node in controller", kind)
        self.purge()
        self.unknown = unknown
        self.controls = controls
        await self.driver.finish_update()
        await asyncio.sleep(0)


RENDERER_CLASS = Controller



"""
Flitter offscreen OpenGL rendering

This is actually all managed within the `window` package, but we need a
module at the top-level of `flitter.render` to ensure that the auto-import
and renderer searching logic works.
"""

from .window import Window


def Offscreen(**kwargs):
    kwargs['offscreen'] = True
    return Window(**kwargs)


RENDERER_CLASS = Offscreen



# cython: language_level=3, profile=True

import cython


cdef class Vector:
    cdef int length
    cdef tuple objects
    cdef double* numbers
    cdef double[16] _numbers
    cdef unsigned long long _hash

    @staticmethod
    cdef Vector _coerce(object other)

    @staticmethod
    cdef Vector _copy(Vector other)

    @staticmethod
    cdef Vector _compose(list vectors)

    @staticmethod
    cdef Vector _symbol(str symbol)

    cdef int allocate_numbers(self, int n) except -1
    cdef void deallocate_numbers(self) noexcept
    cpdef Vector intern(self)
    cdef void fill_range(self, Vector startv, Vector stopv, Vector stepv)
    cpdef bint isinstance(self, t) noexcept
    cdef bint as_bool(self)
    cdef double as_double(self) noexcept
    cdef str as_string(self)
    cdef unsigned long long hash(self, bint floor_floats)
    cpdef object match(self, int n=?, type t=?, default=?)
    cdef str repr(self)
    cdef Vector neg(self)
    cdef Vector pos(self) noexcept
    cdef Vector abs(self)
    cdef Vector ceil(self)
    cdef Vector floor(self)
    cpdef Vector fract(self)
    cdef Vector add(self, Vector other)
    cpdef Vector mul_add(self, Vector left, Vector right)
    cdef Vector sub(self, Vector other)
    cdef Vector mul(self, Vector other)
    cdef Vector truediv(self, Vector other)
    cdef Vector floordiv(self, Vector other)
    cdef Vector mod(self, Vector other)
    cdef Vector pow(self, Vector other)
    cdef Vector eq(self, Vector other)
    cdef Vector ne(self, Vector other)
    cdef Vector gt(self, Vector other)
    cdef Vector ge(self, Vector other)
    cdef Vector lt(self, Vector other)
    cdef Vector le(self, Vector other)
    cdef int compare(self, Vector other) except -2
    cdef Vector slice(self, Vector index)
    cdef Vector item(self, int i)
    cpdef double squared_sum(self) noexcept
    cpdef Vector normalize(self)
    cpdef Vector dot(self, Vector other)
    cpdef Vector cross(self, Vector other)
    cpdef Vector concat(self, Vector other)


cdef Vector null_
cdef Vector true_
cdef Vector false_
cdef Vector minusone_


cdef class Matrix33(Vector):
    @staticmethod
    cdef Matrix33 _translate(Vector v)

    @staticmethod
    cdef Matrix33 _scale(Vector v)

    @staticmethod
    cdef Matrix33 _rotate(double turns)

    cdef Matrix33 mmul(self, Matrix33 b)
    cdef Vector vmul(self, Vector b)
    cpdef Matrix33 inverse(self)
    cpdef Matrix33 transpose(self)


cdef class Matrix44(Vector):
    @staticmethod
    cdef Matrix44 _project(double xgradient, double ygradient, double near, double far)

    @staticmethod
    cdef Matrix44 _ortho(double aspect_ratio, double width, double near, double far)

    @staticmethod
    cdef Matrix44 _look(Vector from_position, Vector to_position, Vector up_direction)

    @staticmethod
    cdef Matrix44 _translate(Vector v)

    @staticmethod
    cdef Matrix44 _scale(Vector v)

    @staticmethod
    cdef Matrix44 _rotate_x(double turns)

    @staticmethod
    cdef Matrix44 _rotate_y(double turns)

    @staticmethod
    cdef Matrix44 _rotate_z(double turns)

    @staticmethod
    cdef Matrix44 _rotate(Vector v)

    @staticmethod
    cdef Matrix44 _shear_x(Vector v)

    @staticmethod
    cdef Matrix44 _shear_y(Vector v)

    @staticmethod
    cdef Matrix44 _shear_z(Vector v)

    cdef Matrix44 mmul(self, Matrix44 b)
    cdef Vector vmul(self, Vector b)
    cpdef Matrix44 inverse(self)
    cpdef Matrix44 transpose(self)
    cpdef Matrix33 inverse_transpose_matrix33(self)
    cpdef Matrix33 matrix33(self)


cdef class Node:
    cdef readonly str kind
    cdef set _tags
    cdef dict _attributes
    cdef bint _attributes_shared
    cdef tuple _children

    cdef unsigned long long hash(self)
    cpdef Node copy(self)
    cpdef void add_tag(self, str tag)
    cpdef void set_attribute(self, str name, Vector value)
    cpdef void append(self, Node node)
    cdef bint _equals(self, Node other)
    cpdef object get(self, str name, int n=?, type t=?, object default=?)
    cdef Vector get_fvec(self, str name, int n, Vector default)
    cdef double get_float(self, str name, double default)
    cdef long get_int(self, str name, long default)
    cdef bint get_bool(self, str name, bint default)
    cdef str get_str(self, str name, str default)
    cdef void repr(self, list lines, int indent)


cdef class StateDict:
    cdef set _changed_keys
    cdef dict _state

    cdef Vector get_item(self, Vector key)
    cdef void set_item(self, Vector key, Vector value)
    cdef bint contains(self, Vector key)


cdef class Context:
    cdef readonly dict names
    cdef readonly set captures
    cdef readonly dict pragmas
    cdef readonly StateDict state
    cdef readonly Node root
    cdef readonly object path
    cdef readonly Context parent
    cdef readonly set errors
    cdef readonly set logs
    cdef readonly dict references
    cdef readonly object stack
    cdef readonly object lnames



# cython: language_level=3, profile=False

"""
OpenSimplex 2(s) noise generation

Lifted from https://github.com/lmas/opensimplex, which itself was adapted from
https://gist.github.com/KdotJPG/b1270127455a94ac5d19
"""

import cython

from libc.math cimport floor

from .functions cimport uniform, shuffle
from ..model cimport Vector, null_


cdef Vector PERM_RANGE = Vector.range(256)

cdef Vector GRADIENTS2 = Vector([5, 2, 2, 5, -5, 2, -2, 5, 5, -2, 2, -5, -5, -2, -2, -5])
cdef double STRETCH_CONSTANT2 = -0.211324865405187
cdef double SQUISH_CONSTANT2 = 0.366025403784439
cdef double NORM_CONSTANT2 = 47

cdef Vector GRADIENTS3 = Vector([
    -11, 4, 4, -4, 11, 4, -4, 4, 11,
    11, 4, 4, 4, 11, 4, 4, 4, 11,
    -11, -4, 4, -4, -11, 4, -4, -4, 11,
    11, -4, 4, 4, -11, 4, 4, -4, 11,
    -11, 4, -4, -4, 11, -4, -4, 4, -11,
    11, 4, -4, 4, 11, -4, 4, 4, -11,
    -11, -4, -4, -4, -11, -4, -4, -4, -11,
    11, -4, -4, 4, -11, -4, 4, -4, -11,
])
cdef double STRETCH_CONSTANT3 = -1.0 / 6
cdef double SQUISH_CONSTANT3 = 1.0 / 3
cdef double NORM_CONSTANT3 = 103

cdef int MAX_PERM_CACHE_ITEMS = 1000
cdef dict PermCache = {}


cdef inline double extrapolate2(Vector perm, long xsb, long ysb, double dx, double dy) noexcept nogil:
    cdef long index = <long>(perm.numbers[(<long>(perm.numbers[xsb % 256]) + ysb) % 256]) % 8 * 2
    return GRADIENTS2.numbers[index] * dx + GRADIENTS2.numbers[index+1] * dy


cdef inline double extrapolate3(Vector perm, long xsb, long ysb, long zsb, double dx, double dy, double dz) noexcept nogil:
    cdef long index = <long>(perm.numbers[(<long>(perm.numbers[(<long>(perm.numbers[xsb % 256]) + ysb) % 256]) + zsb) % 256]) % 24 * 3
    return GRADIENTS3.numbers[index] * dx + GRADIENTS3.numbers[index+1] * dy + GRADIENTS3.numbers[index+2] * dz


@cython.cdivision(True)
cdef double noise2(Vector perm, double x, double y) noexcept nogil:
    cdef long xsb, ysb, xsv_ext, ysv_ext
    stretch_offset = (x + y) * STRETCH_CONSTANT2
    xs = x + stretch_offset
    ys = y + stretch_offset
    xsb = <long>floor(xs)
    ysb = <long>floor(ys)
    squish_offset = (xsb + ysb) * SQUISH_CONSTANT2
    xb = xsb + squish_offset
    yb = ysb + squish_offset
    xins = xs - xsb
    yins = ys - ysb
    in_sum = xins + yins
    dx0 = x - xb
    dy0 = y - yb
    value = 0.0
    dx1 = dx0 - 1 - SQUISH_CONSTANT2
    dy1 = dy0 - 0 - SQUISH_CONSTANT2
    attn1 = 2 - dx1 * dx1 - dy1 * dy1
    if attn1 > 0:
        attn1 *= attn1
        value += attn1 * attn1 * extrapolate2(perm, xsb + 1, ysb + 0, dx1, dy1)
    dx2 = dx0 - 0 - SQUISH_CONSTANT2
    dy2 = dy0 - 1 - SQUISH_CONSTANT2
    attn2 = 2 - dx2 * dx2 - dy2 * dy2
    if attn2 > 0:
        attn2 *= attn2
        value += attn2 * attn2 * extrapolate2(perm, xsb + 0, ysb + 1, dx2, dy2)
    if in_sum <= 1:
        zins = 1 - in_sum
        if zins > xins or zins > yins:
            if xins > yins:
                xsv_ext = xsb + 1
                ysv_ext = ysb - 1
                dx_ext = dx0 - 1
                dy_ext = dy0 + 1
            else:
                xsv_ext = xsb - 1
                ysv_ext = ysb + 1
                dx_ext = dx0 + 1
                dy_ext = dy0 - 1
        else:
            xsv_ext = xsb + 1
            ysv_ext = ysb + 1
            dx_ext = dx0 - 1 - 2 * SQUISH_CONSTANT2
            dy_ext = dy0 - 1 - 2 * SQUISH_CONSTANT2
    else:
        zins = 2 - in_sum
        if zins < xins or zins < yins:
            if xins > yins:
                xsv_ext = xsb + 2
                ysv_ext = ysb + 0
                dx_ext = dx0 - 2 - 2 * SQUISH_CONSTANT2
                dy_ext = dy0 + 0 - 2 * SQUISH_CONSTANT2
            else:
                xsv_ext = xsb + 0
                ysv_ext = ysb + 2
                dx_ext = dx0 + 0 - 2 * SQUISH_CONSTANT2
                dy_ext = dy0 - 2 - 2 * SQUISH_CONSTANT2
        else:
            dx_ext = dx0
            dy_ext = dy0
            xsv_ext = xsb
            ysv_ext = ysb
        xsb += 1
        ysb += 1
        dx0 = dx0 - 1 - 2 * SQUISH_CONSTANT2
        dy0 = dy0 - 1 - 2 * SQUISH_CONSTANT2
    attn0 = 2 - dx0 * dx0 - dy0 * dy0
    if attn0 > 0:
        attn0 *= attn0
        value += attn0 * attn0 * extrapolate2(perm, xsb, ysb, dx0, dy0)
    attn_ext = 2 - dx_ext * dx_ext - dy_ext * dy_ext
    if attn_ext > 0:
        attn_ext *= attn_ext
        value += attn_ext * attn_ext * extrapolate2(perm, xsv_ext, ysv_ext, dx_ext, dy_ext)
    return value / NORM_CONSTANT2


@cython.cdivision(True)
cdef double noise3(Vector perm, double x, double y, double z) noexcept nogil:
    cdef long xsb, ysb, zsb, xsv_ext0, xsv_ext1, ysv_ext0, ysv_ext1, zsv_ext0, zsv_ext1
    stretch_offset = (x + y + z) * STRETCH_CONSTANT3
    xs = x + stretch_offset
    ys = y + stretch_offset
    zs = z + stretch_offset
    xsb = <long>floor(xs)
    ysb = <long>floor(ys)
    zsb = <long>floor(zs)
    squish_offset = (xsb + ysb + zsb) * SQUISH_CONSTANT3
    xb = xsb + squish_offset
    yb = ysb + squish_offset
    zb = zsb + squish_offset
    xins = xs - xsb
    yins = ys - ysb
    zins = zs - zsb
    in_sum = xins + yins + zins
    dx0 = x - xb
    dy0 = y - yb
    dz0 = z - zb
    value = 0.0
    if in_sum <= 1:
        a_point = 0x01
        a_score = xins
        b_point = 0x02
        b_score = yins
        if a_score >= b_score and zins > b_score:
            b_score = zins
            b_point = 0x04
        elif a_score < b_score and zins > a_score:
            a_score = zins
            a_point = 0x04
        wins = 1 - in_sum
        if wins > a_score or wins > b_score:
            c = b_point if (b_score > a_score) else a_point
            if (c & 0x01) == 0:
                xsv_ext0 = xsb - 1
                xsv_ext1 = xsb
                dx_ext0 = dx0 + 1
                dx_ext1 = dx0
            else:
                xsv_ext0 = xsv_ext1 = xsb + 1
                dx_ext0 = dx_ext1 = dx0 - 1
            if (c & 0x02) == 0:
                ysv_ext0 = ysv_ext1 = ysb
                dy_ext0 = dy_ext1 = dy0
                if (c & 0x01) == 0:
                    ysv_ext1 -= 1
                    dy_ext1 += 1
                else:
                    ysv_ext0 -= 1
                    dy_ext0 += 1
            else:
                ysv_ext0 = ysv_ext1 = ysb + 1
                dy_ext0 = dy_ext1 = dy0 - 1
            if (c & 0x04) == 0:
                zsv_ext0 = zsb
                zsv_ext1 = zsb - 1
                dz_ext0 = dz0
                dz_ext1 = dz0 + 1
            else:
                zsv_ext0 = zsv_ext1 = zsb + 1
                dz_ext0 = dz_ext1 = dz0 - 1
        else:
            c = a_point | b_point
            if (c & 0x01) == 0:
                xsv_ext0 = xsb
                xsv_ext1 = xsb - 1
                dx_ext0 = dx0 - 2 * SQUISH_CONSTANT3
                dx_ext1 = dx0 + 1 - SQUISH_CONSTANT3
            else:
                xsv_ext0 = xsv_ext1 = xsb + 1
                dx_ext0 = dx0 - 1 - 2 * SQUISH_CONSTANT3
                dx_ext1 = dx0 - 1 - SQUISH_CONSTANT3
            if (c & 0x02) == 0:
                ysv_ext0 = ysb
                ysv_ext1 = ysb - 1
                dy_ext0 = dy0 - 2 * SQUISH_CONSTANT3
                dy_ext1 = dy0 + 1 - SQUISH_CONSTANT3
            else:
                ysv_ext0 = ysv_ext1 = ysb + 1
                dy_ext0 = dy0 - 1 - 2 * SQUISH_CONSTANT3
                dy_ext1 = dy0 - 1 - SQUISH_CONSTANT3
            if (c & 0x04) == 0:
                zsv_ext0 = zsb
                zsv_ext1 = zsb - 1
                dz_ext0 = dz0 - 2 * SQUISH_CONSTANT3
                dz_ext1 = dz0 + 1 - SQUISH_CONSTANT3
            else:
                zsv_ext0 = zsv_ext1 = zsb + 1
                dz_ext0 = dz0 - 1 - 2 * SQUISH_CONSTANT3
                dz_ext1 = dz0 - 1 - SQUISH_CONSTANT3
        attn0 = 2 - dx0 * dx0 - dy0 * dy0 - dz0 * dz0
        if attn0 > 0:
            attn0 *= attn0
            value += attn0 * attn0 * extrapolate3(perm, xsb + 0, ysb + 0, zsb + 0, dx0, dy0, dz0)
        dx1 = dx0 - 1 - SQUISH_CONSTANT3
        dy1 = dy0 - 0 - SQUISH_CONSTANT3
        dz1 = dz0 - 0 - SQUISH_CONSTANT3
        attn1 = 2 - dx1 * dx1 - dy1 * dy1 - dz1 * dz1
        if attn1 > 0:
            attn1 *= attn1
            value += attn1 * attn1 * extrapolate3(perm, xsb + 1, ysb + 0, zsb + 0, dx1, dy1, dz1)
        dx2 = dx0 - 0 - SQUISH_CONSTANT3
        dy2 = dy0 - 1 - SQUISH_CONSTANT3
        dz2 = dz1
        attn2 = 2 - dx2 * dx2 - dy2 * dy2 - dz2 * dz2
        if attn2 > 0:
            attn2 *= attn2
            value += attn2 * attn2 * extrapolate3(perm, xsb + 0, ysb + 1, zsb + 0, dx2, dy2, dz2)
        dx3 = dx2
        dy3 = dy1
        dz3 = dz0 - 1 - SQUISH_CONSTANT3
        attn3 = 2 - dx3 * dx3 - dy3 * dy3 - dz3 * dz3
        if attn3 > 0:
            attn3 *= attn3
            value += attn3 * attn3 * extrapolate3(perm, xsb + 0, ysb + 0, zsb + 1, dx3, dy3, dz3)
    elif in_sum >= 2:
        a_point = 0x06
        a_score = xins
        b_point = 0x05
        b_score = yins
        if a_score <= b_score and zins < b_score:
            b_score = zins
            b_point = 0x03
        elif a_score > b_score and zins < a_score:
            a_score = zins
            a_point = 0x03
        wins = 3 - in_sum
        if wins < a_score or wins < b_score:
            c = b_point if (b_score < a_score) else a_point
            if (c & 0x01) != 0:
                xsv_ext0 = xsb + 2
                xsv_ext1 = xsb + 1
                dx_ext0 = dx0 - 2 - 3 * SQUISH_CONSTANT3
                dx_ext1 = dx0 - 1 - 3 * SQUISH_CONSTANT3
            else:
                xsv_ext0 = xsv_ext1 = xsb
                dx_ext0 = dx_ext1 = dx0 - 3 * SQUISH_CONSTANT3
            if (c & 0x02) != 0:
                ysv_ext0 = ysv_ext1 = ysb + 1
                dy_ext0 = dy_ext1 = dy0 - 1 - 3 * SQUISH_CONSTANT3
                if (c & 0x01) != 0:
                    ysv_ext1 += 1
                    dy_ext1 -= 1
                else:
                    ysv_ext0 += 1
                    dy_ext0 -= 1
            else:
                ysv_ext0 = ysv_ext1 = ysb
                dy_ext0 = dy_ext1 = dy0 - 3 * SQUISH_CONSTANT3
            if (c & 0x04) != 0:
                zsv_ext0 = zsb + 1
                zsv_ext1 = zsb + 2
                dz_ext0 = dz0 - 1 - 3 * SQUISH_CONSTANT3
                dz_ext1 = dz0 - 2 - 3 * SQUISH_CONSTANT3
            else:
                zsv_ext0 = zsv_ext1 = zsb
                dz_ext0 = dz_ext1 = dz0 - 3 * SQUISH_CONSTANT3
        else:
            c = a_point & b_point
            if (c & 0x01) != 0:
                xsv_ext0 = xsb + 1
                xsv_ext1 = xsb + 2
                dx_ext0 = dx0 - 1 - SQUISH_CONSTANT3
                dx_ext1 = dx0 - 2 - 2 * SQUISH_CONSTANT3
            else:
                xsv_ext0 = xsv_ext1 = xsb
                dx_ext0 = dx0 - SQUISH_CONSTANT3
                dx_ext1 = dx0 - 2 * SQUISH_CONSTANT3
            if (c & 0x02) != 0:
                ysv_ext0 = ysb + 1
                ysv_ext1 = ysb + 2
                dy_ext0 = dy0 - 1 - SQUISH_CONSTANT3
                dy_ext1 = dy0 - 2 - 2 * SQUISH_CONSTANT3
            else:
                ysv_ext0 = ysv_ext1 = ysb
                dy_ext0 = dy0 - SQUISH_CONSTANT3
                dy_ext1 = dy0 - 2 * SQUISH_CONSTANT3
            if (c & 0x04) != 0:
                zsv_ext0 = zsb + 1
                zsv_ext1 = zsb + 2
                dz_ext0 = dz0 - 1 - SQUISH_CONSTANT3
                dz_ext1 = dz0 - 2 - 2 * SQUISH_CONSTANT3
            else:
                zsv_ext0 = zsv_ext1 = zsb
                dz_ext0 = dz0 - SQUISH_CONSTANT3
                dz_ext1 = dz0 - 2 * SQUISH_CONSTANT3
        dx3 = dx0 - 1 - 2 * SQUISH_CONSTANT3
        dy3 = dy0 - 1 - 2 * SQUISH_CONSTANT3
        dz3 = dz0 - 0 - 2 * SQUISH_CONSTANT3
        attn3 = 2 - dx3 * dx3 - dy3 * dy3 - dz3 * dz3
        if attn3 > 0:
            attn3 *= attn3
            value += attn3 * attn3 * extrapolate3(perm, xsb + 1, ysb + 1, zsb + 0, dx3, dy3, dz3)
        dx2 = dx3
        dy2 = dy0 - 0 - 2 * SQUISH_CONSTANT3
        dz2 = dz0 - 1 - 2 * SQUISH_CONSTANT3
        attn2 = 2 - dx2 * dx2 - dy2 * dy2 - dz2 * dz2
        if attn2 > 0:
            attn2 *= attn2
            value += attn2 * attn2 * extrapolate3(perm, xsb + 1, ysb + 0, zsb + 1, dx2, dy2, dz2)
        dx1 = dx0 - 0 - 2 * SQUISH_CONSTANT3
        dy1 = dy3
        dz1 = dz2
        attn1 = 2 - dx1 * dx1 - dy1 * dy1 - dz1 * dz1
        if attn1 > 0:
            attn1 *= attn1
            value += attn1 * attn1 * extrapolate3(perm, xsb + 0, ysb + 1, zsb + 1, dx1, dy1, dz1)
        dx0 = dx0 - 1 - 3 * SQUISH_CONSTANT3
        dy0 = dy0 - 1 - 3 * SQUISH_CONSTANT3
        dz0 = dz0 - 1 - 3 * SQUISH_CONSTANT3
        attn0 = 2 - dx0 * dx0 - dy0 * dy0 - dz0 * dz0
        if attn0 > 0:
            attn0 *= attn0
            value += attn0 * attn0 * extrapolate3(perm, xsb + 1, ysb + 1, zsb + 1, dx0, dy0, dz0)
    else:
        p1 = xins + yins
        if p1 > 1:
            a_score = p1 - 1
            a_point = 0x03
            a_is_further_side = True
        else:
            a_score = 1 - p1
            a_point = 0x04
            a_is_further_side = False
        p2 = xins + zins
        if p2 > 1:
            b_score = p2 - 1
            b_point = 0x05
            b_is_further_side = True
        else:
            b_score = 1 - p2
            b_point = 0x02
            b_is_further_side = False
        p3 = yins + zins
        if p3 > 1:
            score = p3 - 1
            if a_score <= b_score and a_score < score:
                a_point = 0x06
                a_is_further_side = True
            elif a_score > b_score and b_score < score:
                b_point = 0x06
                b_is_further_side = True
        else:
            score = 1 - p3
            if a_score <= b_score and a_score < score:
                a_point = 0x01
                a_is_further_side = False
            elif a_score > b_score and b_score < score:
                b_point = 0x01
                b_is_further_side = False
        if a_is_further_side == b_is_further_side:
            if a_is_further_side:
                dx_ext0 = dx0 - 1 - 3 * SQUISH_CONSTANT3
                dy_ext0 = dy0 - 1 - 3 * SQUISH_CONSTANT3
                dz_ext0 = dz0 - 1 - 3 * SQUISH_CONSTANT3
                xsv_ext0 = xsb + 1
                ysv_ext0 = ysb + 1
                zsv_ext0 = zsb + 1
                c = a_point & b_point
                if (c & 0x01) != 0:
                    dx_ext1 = dx0 - 2 - 2 * SQUISH_CONSTANT3
                    dy_ext1 = dy0 - 2 * SQUISH_CONSTANT3
                    dz_ext1 = dz0 - 2 * SQUISH_CONSTANT3
                    xsv_ext1 = xsb + 2
                    ysv_ext1 = ysb
                    zsv_ext1 = zsb
                elif (c & 0x02) != 0:
                    dx_ext1 = dx0 - 2 * SQUISH_CONSTANT3
                    dy_ext1 = dy0 - 2 - 2 * SQUISH_CONSTANT3
                    dz_ext1 = dz0 - 2 * SQUISH_CONSTANT3
                    xsv_ext1 = xsb
                    ysv_ext1 = ysb + 2
                    zsv_ext1 = zsb
                else:
                    dx_ext1 = dx0 - 2 * SQUISH_CONSTANT3
                    dy_ext1 = dy0 - 2 * SQUISH_CONSTANT3
                    dz_ext1 = dz0 - 2 - 2 * SQUISH_CONSTANT3
                    xsv_ext1 = xsb
                    ysv_ext1 = ysb
                    zsv_ext1 = zsb + 2
            else:
                dx_ext0 = dx0
                dy_ext0 = dy0
                dz_ext0 = dz0
                xsv_ext0 = xsb
                ysv_ext0 = ysb
                zsv_ext0 = zsb
                c = a_point | b_point
                if (c & 0x01) == 0:
                    dx_ext1 = dx0 + 1 - SQUISH_CONSTANT3
                    dy_ext1 = dy0 - 1 - SQUISH_CONSTANT3
                    dz_ext1 = dz0 - 1 - SQUISH_CONSTANT3
                    xsv_ext1 = xsb - 1
                    ysv_ext1 = ysb + 1
                    zsv_ext1 = zsb + 1
                elif (c & 0x02) == 0:
                    dx_ext1 = dx0 - 1 - SQUISH_CONSTANT3
                    dy_ext1 = dy0 + 1 - SQUISH_CONSTANT3
                    dz_ext1 = dz0 - 1 - SQUISH_CONSTANT3
                    xsv_ext1 = xsb + 1
                    ysv_ext1 = ysb - 1
                    zsv_ext1 = zsb + 1
                else:
                    dx_ext1 = dx0 - 1 - SQUISH_CONSTANT3
                    dy_ext1 = dy0 - 1 - SQUISH_CONSTANT3
                    dz_ext1 = dz0 + 1 - SQUISH_CONSTANT3
                    xsv_ext1 = xsb + 1
                    ysv_ext1 = ysb + 1
                    zsv_ext1 = zsb - 1
        else:
            if a_is_further_side:
                c1 = a_point
                c2 = b_point
            else:
                c1 = b_point
                c2 = a_point
            if (c1 & 0x01) == 0:
                dx_ext0 = dx0 + 1 - SQUISH_CONSTANT3
                dy_ext0 = dy0 - 1 - SQUISH_CONSTANT3
                dz_ext0 = dz0 - 1 - SQUISH_CONSTANT3
                xsv_ext0 = xsb - 1
                ysv_ext0 = ysb + 1
                zsv_ext0 = zsb + 1
            elif (c1 & 0x02) == 0:
                dx_ext0 = dx0 - 1 - SQUISH_CONSTANT3
                dy_ext0 = dy0 + 1 - SQUISH_CONSTANT3
                dz_ext0 = dz0 - 1 - SQUISH_CONSTANT3
                xsv_ext0 = xsb + 1
                ysv_ext0 = ysb - 1
                zsv_ext0 = zsb + 1
            else:
                dx_ext0 = dx0 - 1 - SQUISH_CONSTANT3
                dy_ext0 = dy0 - 1 - SQUISH_CONSTANT3
                dz_ext0 = dz0 + 1 - SQUISH_CONSTANT3
                xsv_ext0 = xsb + 1
                ysv_ext0 = ysb + 1
                zsv_ext0 = zsb - 1
            dx_ext1 = dx0 - 2 * SQUISH_CONSTANT3
            dy_ext1 = dy0 - 2 * SQUISH_CONSTANT3
            dz_ext1 = dz0 - 2 * SQUISH_CONSTANT3
            xsv_ext1 = xsb
            ysv_ext1 = ysb
            zsv_ext1 = zsb
            if (c2 & 0x01) != 0:
                dx_ext1 -= 2
                xsv_ext1 += 2
            elif (c2 & 0x02) != 0:
                dy_ext1 -= 2
                ysv_ext1 += 2
            else:
                dz_ext1 -= 2
                zsv_ext1 += 2
        dx1 = dx0 - 1 - SQUISH_CONSTANT3
        dy1 = dy0 - 0 - SQUISH_CONSTANT3
        dz1 = dz0 - 0 - SQUISH_CONSTANT3
        attn1 = 2 - dx1 * dx1 - dy1 * dy1 - dz1 * dz1
        if attn1 > 0:
            attn1 *= attn1
            value += attn1 * attn1 * extrapolate3(perm, xsb + 1, ysb + 0, zsb + 0, dx1, dy1, dz1)
        dx2 = dx0 - 0 - SQUISH_CONSTANT3
        dy2 = dy0 - 1 - SQUISH_CONSTANT3
        dz2 = dz1
        attn2 = 2 - dx2 * dx2 - dy2 * dy2 - dz2 * dz2
        if attn2 > 0:
            attn2 *= attn2
            value += attn2 * attn2 * extrapolate3(perm, xsb + 0, ysb + 1, zsb + 0, dx2, dy2, dz2)
        dx3 = dx2
        dy3 = dy1
        dz3 = dz0 - 1 - SQUISH_CONSTANT3
        attn3 = 2 - dx3 * dx3 - dy3 * dy3 - dz3 * dz3
        if attn3 > 0:
            attn3 *= attn3
            value += attn3 * attn3 * extrapolate3(perm, xsb + 0, ysb + 0, zsb + 1, dx3, dy3, dz3)
        dx4 = dx0 - 1 - 2 * SQUISH_CONSTANT3
        dy4 = dy0 - 1 - 2 * SQUISH_CONSTANT3
        dz4 = dz0 - 0 - 2 * SQUISH_CONSTANT3
        attn4 = 2 - dx4 * dx4 - dy4 * dy4 - dz4 * dz4
        if attn4 > 0:
            attn4 *= attn4
            value += attn4 * attn4 * extrapolate3(perm, xsb + 1, ysb + 1, zsb + 0, dx4, dy4, dz4)
        dx5 = dx4
        dy5 = dy0 - 0 - 2 * SQUISH_CONSTANT3
        dz5 = dz0 - 1 - 2 * SQUISH_CONSTANT3
        attn5 = 2 - dx5 * dx5 - dy5 * dy5 - dz5 * dz5
        if attn5 > 0:
            attn5 *= attn5
            value += attn5 * attn5 * extrapolate3(perm, xsb + 1, ysb + 0, zsb + 1, dx5, dy5, dz5)
        dx6 = dx0 - 0 - 2 * SQUISH_CONSTANT3
        dy6 = dy4
        dz6 = dz5
        attn6 = 2 - dx6 * dx6 - dy6 * dy6 - dz6 * dz6
        if attn6 > 0:
            attn6 *= attn6
            value += attn6 * attn6 * extrapolate3(perm, xsb + 0, ysb + 1, zsb + 1, dx6, dy6, dz6)
    attn_ext0 = 2 - dx_ext0 * dx_ext0 - dy_ext0 * dy_ext0 - dz_ext0 * dz_ext0
    if attn_ext0 > 0:
        attn_ext0 *= attn_ext0
        value += attn_ext0 * attn_ext0 * extrapolate3(perm, xsv_ext0, ysv_ext0, zsv_ext0, dx_ext0, dy_ext0, dz_ext0)
    attn_ext1 = 2 - dx_ext1 * dx_ext1 - dy_ext1 * dy_ext1 - dz_ext1 * dz_ext1
    if attn_ext1 > 0:
        attn_ext1 *= attn_ext1
        value += attn_ext1 * attn_ext1 * extrapolate3(perm, xsv_ext1, ysv_ext1, zsv_ext1, dx_ext1, dy_ext1, dz_ext1)
    return value / NORM_CONSTANT3


cdef Vector _noise(Vector perm, list args):
    cdef int i, j, k, count = 0, n = len(args)
    cdef Vector xx, yy, zz
    cdef double x, y
    cdef Vector result = Vector.__new__(Vector)
    if n == 1:
        xx = args[0]
        if xx.numbers != NULL:
            result.allocate_numbers(xx.length)
            with nogil:
                for i in range(xx.length):
                    result.numbers[i] = noise2(perm, xx.numbers[i], 0)
    elif n == 2:
        xx = args[0]
        yy = args[1]
        if xx.numbers != NULL and yy.numbers != NULL:
            result.allocate_numbers(xx.length * yy.length)
            with nogil:
                for i in range(xx.length):
                    x = xx.numbers[i]
                    for j in range(yy.length):
                        result.numbers[count] = noise2(perm, x, yy.numbers[j])
                        count += 1
    elif n == 3:
        xx = args[0]
        yy = args[1]
        zz = args[2]
        if xx.numbers != NULL and yy.numbers != NULL and zz.numbers != NULL:
            result.allocate_numbers(xx.length * yy.length * zz.length)
            with nogil:
                for i in range(xx.length):
                    x = xx.numbers[i]
                    for j in range(yy.length):
                        y = yy.numbers[j]
                        for k in range(zz.length):
                            result.numbers[count] = noise3(perm, x, y, zz.numbers[k])
                            count += 1
    return result


cdef Vector get_perm(Vector seed, int i):
    cdef unsigned long long seed_hash = seed.hash(True) ^ <unsigned long long>i
    cdef uniform prng
    cdef Vector perm = <Vector>PermCache.get(seed_hash)
    if perm is None:
        prng = uniform.__new__(uniform)
        prng._hash = seed_hash
        perm = shuffle(prng, PERM_RANGE)
        if len(PermCache) == MAX_PERM_CACHE_ITEMS:
            PermCache.pop(next(iter(PermCache)))
        PermCache[seed_hash] = perm
    return perm


def noise(Vector seed, *args):
    return _noise(get_perm(seed, 0), list(args))


@cython.cdivision(True)
def octnoise(Vector seed, Vector octaves, Vector roughness, *args):
    if octaves.numbers == NULL or octaves.length != 1 or roughness.numbers == NULL or roughness.length != 1:
        return null_
    cdef Vector arg
    cdef list coords = []
    for arg in args:
        if arg.numbers == NULL:
            return null_
        coords.append(Vector._copy(arg))
    cdef int i, j, n = <int>octaves.numbers[0]
    cdef double weight_sum = 0, weight = 1, k = roughness.numbers[0]
    cdef Vector single, result = null_
    for i in range(n):
        if i == 0:
            result = _noise(get_perm(seed, i), coords)
        else:
            single = _noise(get_perm(seed, i), coords)
            for j in range(result.length):
                result.numbers[j] += single.numbers[j] * weight
        for arg in coords:
            for j in range(arg.length):
                arg.numbers[j] *= 2
        weight_sum += weight
        weight *= k
    for j in range(result.length):
        result.numbers[j] /= weight_sum
    return result


NOISE_FUNCTIONS = {
    'noise': Vector(noise),
    'octnoise': Vector(octnoise),
}



"""
General file cache
"""

import fractions
from os import stat
from pathlib import Path
from queue import Queue, Full
import threading

from loguru import logger

from .clock import system_clock


DEFAULT_CLEAN_TIME = 30
MAX_CACHE_VIDEO_FRAMES = 100


class CachePath:
    def __init__(self, path, absolute):
        self._path = path
        self._absolute = absolute
        self._cache = {}
        self._mtime = None
        self.check_unmodified()

    @property
    def suffix(self):
        return self._path.suffix

    def exists(self):
        return self._path.exists()

    def cleanup(self):
        while self._cache:
            key, value = self._cache.popitem()
            match key:
                case 'video', _:
                    container, decoder, frames = value
                    if decoder is not None:
                        decoder.close()
                    if container is not None:
                        container.close()
                        logger.debug("Closing video: {}", self._path)
                case 'video_output':
                    writer, queue, *_ = value
                    if queue is not None:
                        queue.put(None)
                        writer.join()

    def check_unmodified(self):
        self._touched = system_clock()
        try:
            mtime = stat(self._absolute).st_mtime
        except FileNotFoundError:
            mtime = None
        if mtime == self._mtime:
            return True
        self._mtime = mtime
        self._cache = {}
        return False

    def read_text(self, encoding=None, errors=None):
        key = 'text', encoding, errors
        if self.check_unmodified() and (text := self._cache.get(key, False)) is not False:
            return text
        if self._mtime is None:
            logger.warning("File not found: {}", self._path)
            text = None
        else:
            try:
                text = self._path.read_text(encoding, errors)
            except Exception as exc:
                logger.opt(exception=exc).warning("Error reading text: {}", self._path)
                text = None
            else:
                logger.debug("Read text: {}", self._path)
        self._cache[key] = text
        return text

    def read_bytes(self):
        key = 'bytes'
        if self.check_unmodified() and (data := self._cache.get(key, False)) is not False:
            return data
        if self._mtime is None:
            logger.warning("File not found: {}", self._path)
            data = None
        else:
            try:
                data = self._path.read_bytes()
            except Exception as exc:
                logger.opt(exception=exc).warning("Error reading bytes: {}", self._path)
                data = None
            else:
                logger.debug("Read bytes: {}", self._path)
        self._cache[key] = data
        return data

    def read_flitter_program(self, static=None, dynamic=None):
        key = 'flitter', tuple(sorted(static.items())) if static else None, tuple(dynamic) if dynamic else ()
        current_program = self._cache.get(key, False)
        if current_program is not False and self.check_unmodified():
            return current_program
        if self._mtime is None:
            if current_program is False:
                logger.warning("Program not found: {}", self._path)
                program = current_program = None
            elif current_program is not None:
                logger.error("Program disappeared: {}", self._path)
                program = current_program
        else:
            if current_program is False:
                current_program = None
            from .language.parser import parse, ParseError
            try:
                parse_time = -system_clock()
                source = self._path.read_text(encoding='utf8')
                initial_top = parse(source)
                now = system_clock()
                parse_time += now
                simplify_time = -now
                top = initial_top.simplify(static=static, dynamic=dynamic)
                now = system_clock()
                simplify_time += now
                compile_time = -now
                program = top.compile(initial_lnames=tuple(dynamic) if dynamic else ())
                program.set_top(top)
                program.set_path(self)
                compile_time += system_clock()
                logger.debug("Read program: {}", self._path)
                logger.debug("Compiled to {} instructions in {:.1f}/{:.1f}/{:.1f}ms",
                             len(program), parse_time*1000, simplify_time*1000, compile_time*1000)
            except ParseError as exc:
                if current_program is None:
                    logger.error("Error parsing {} at line {} column {}:\n{}",
                                 self._path.name, exc.line, exc.column, exc.context)
                else:
                    logger.warning("Unable to re-parse {}, error at line {} column {}:\n{}",
                                   self._path.name, exc.line, exc.column, exc.context)
                program = current_program
            except Exception as exc:
                logger.opt(exception=exc).error("Error reading program: {}", self._path)
                program = current_program
        self._cache[key] = program
        return program

    def read_csv_vector(self, row_number):
        import csv
        from .model import Vector, null
        if self.check_unmodified() and (cached := self._cache.get('csv', False)) is not False:
            reader, rows = cached
        elif self._mtime is None:
            logger.warning("File not found: {}", self._path)
            reader = None
            rows = []
        else:
            try:
                reader = csv.reader(self._path.open(newline='', encoding='utf8'))
                logger.debug("Opened CSV file: {}", self._path)
            except Exception as exc:
                logger.opt(exception=exc).warning("Error reading CSV: {}", self._path)
                reader = None
            rows = []
        while reader is not None and row_number >= len(rows):
            try:
                row = next(reader)
                values = []
                for value in row:
                    try:
                        value = float(value)
                    except ValueError:
                        pass
                    values.append(value)
                while values and values[-1] == '':
                    values.pop()
                rows.append(Vector.coerce(values))
            except StopIteration:
                logger.debug("Closed CSV file: {}", self._path)
                reader = None
                break
            except Exception as exc:
                logger.opt(exception=exc).warning("Error reading CSV: {}", self._path)
                reader = None
        self._cache['csv'] = reader, rows
        if 0 <= row_number < len(rows):
            return rows[row_number]
        return null

    def read_pil_image(self):
        if self.check_unmodified() and (image := self._cache.get('pil_image', False)) is not False:
            return image
        import PIL.Image
        if self._mtime is None:
            logger.warning("File not found: {}", self._path)
            image = None
        else:
            try:
                image = PIL.Image.open(str(self._path))
            except Exception as exc:
                logger.opt(exception=exc).warning("Error reading image: {}", self._path)
                image = None
            else:
                logger.debug("Read image file: {}", self._path)
        self._cache['pil_image'] = image
        return image

    def read_skia_image(self):
        if self.check_unmodified() and (image := self._cache.get('skia_image', False)) is not False:
            return image
        import skia
        if self._mtime is None:
            logger.warning("File not found: {}", self._path)
            image = None
        else:
            try:
                image = skia.Image.open(str(self._path))
            except Exception as exc:
                logger.opt(exception=exc).warning("Error reading image: {}", self._path)
                image = None
            else:
                logger.debug("Read image file: {}", self._path)
        self._cache['skia_image'] = image
        return image

    def read_video_frames(self, obj, position, loop=False, threading=False):
        key = 'video', id(obj)
        container = decoder = current_frame = next_frame = None
        frames = []
        ratio = 0
        if self.check_unmodified() and (cached := self._cache.get(key, False)) is not False:
            container, decoder, frames = cached
        elif self._mtime is None:
            logger.warning("File not found: {}", self._path)
        else:
            try:
                import av
                container = av.container.open(str(self._path))
                stream = container.streams.video[0]
                if threading:
                    stream.thread_type = 'AUTO'
                ctx = stream.codec_context
                logger.debug("Opened {}x{} {:.0f}fps video file: {}", ctx.width, ctx.height, float(stream.average_rate), self._path)
            except Exception as exc:
                logger.opt(exception=exc).warning("Error reading video: {}", self._path)
                container = None
        if container is not None:
            try:
                stream = container.streams.video[0]
                if loop:
                    timestamp = stream.start_time + int(position / stream.time_base) % stream.duration
                else:
                    timestamp = stream.start_time + min(max(0, int(position / stream.time_base)), stream.duration)
                count = 0
                while True:
                    if len(frames) >= 2 and timestamp >= frames[0].pts and (frames[-1] is None or timestamp < frames[-1].pts):
                        break
                    if frames and timestamp < frames[0].pts:
                        logger.trace("Discard {} buffered video frames", len(frames))
                        frames = []
                    if len(frames) >= 2 and frames[-1].key_frame and timestamp > 2*frames[-1].pts - frames[0].pts:
                        logger.trace("Discard {} buffered video frames", len(frames))
                        frames = []
                    if not frames:
                        if decoder is not None:
                            decoder.close()
                        logger.trace("Seek video {} to {:.2f}s", self._path, float(timestamp * stream.time_base))
                        container.seek(timestamp, stream=stream)
                        decoder = container.decode(streams=(stream.index,))
                    if decoder is not None:
                        try:
                            frame = next(decoder)
                            if frame.key_frame:
                                logger.trace("Read video key frame @ {:.2f}s", float(frame.pts * stream.time_base))
                                for i in range(len(frames)-1, 0, -1):
                                    if frames[i].key_frame:
                                        frames = frames[i:]
                                        logger.trace("Discard {} buffered video frames", i)
                                        break
                            else:
                                logger.trace("Read video frame @ {:.2f}s", float(frame.pts * stream.time_base))
                            count += 1
                        except StopIteration:
                            logger.trace("Hit end of video {}", self._path)
                            decoder = None
                            frame = None
                        if len(frames) == MAX_CACHE_VIDEO_FRAMES:
                            frames.pop(0)
                            logger.trace("Discard one buffered video frame (hit maximum)")
                        frames.append(frame)
                        if len(frames) == 1:
                            logger.trace("Decoding frames from {:.2f}s", float(frames[0].pts * stream.time_base))
                if count > 1:
                    logger.trace("Decoded {} frames to find position {:.2f}s", count, float(timestamp * stream.time_base))
                for current_frame in reversed(frames):
                    if current_frame is not None and current_frame.pts <= timestamp:
                        break
                    next_frame = current_frame
                ratio = 0 if next_frame is None else (timestamp - current_frame.pts) / (next_frame.pts - current_frame.pts)
            except Exception as exc:
                logger.opt(exception=exc).warning("Error reading video: {}", self._path)
                container = decoder = None
                frames = []
        self._cache[key] = container, decoder, frames
        return ratio, current_frame, next_frame

    def read_trimesh_model(self):
        if self.check_unmodified() and (trimesh_model := self._cache.get('trimesh', False)) is not False:
            return trimesh_model
        if self._mtime is None:
            logger.warning("File not found: {}", self._path)
            trimesh_model = None
        else:
            try:
                import trimesh
                trimesh_model = trimesh.load(str(self._path))
            except Exception as exc:
                logger.opt(exception=exc).warning("Error reading model file: {}", self._path)
                trimesh_model = None
            else:
                logger.debug("Read {}v/{}f mesh: {}", len(trimesh_model.vertices), len(trimesh_model.faces), self._path)
        self._cache['trimesh'] = trimesh_model
        return trimesh_model

    def write_image(self, texture, quality=None):
        import PIL.Image
        import PIL.ImageCms
        self._touched = system_clock()
        if 'write_image' in self._cache:
            return
        suffix = self._path.suffix.lower()
        registered_extensions = PIL.Image.registered_extensions()
        if suffix not in registered_extensions:
            logger.warning("Unrecognised image suffix: {}", self._path)
        else:
            self.cleanup()
            if self._path.exists():
                logger.warning("Existing image file will be overwritten: {}", self._path)
            image = PIL.Image.frombytes('RGBA' if texture.components == 4 else 'RGB', (texture.width, texture.height), texture.read())
            encoder = registered_extensions[suffix]
            if image.mode != 'RGB' and encoder not in ('PNG', 'TIFF', 'GIF', 'JPEG2000', 'WEBP'):
                image = image.convert('RGB')
            options = {'icc_profile': PIL.ImageCms.ImageCmsProfile(PIL.ImageCms.createProfile('sRGB')).tobytes()}
            if quality:
                options['quality'] = quality
            try:
                image.save(self._path, **options)
            except Exception as exc:
                logger.opt(exception=exc).error("Unable to save image to: {}", self._path)
            else:
                logger.success("Saved image to file: {}", self._path)
        self._cache['write_image'] = True

    def write_video_frame(self, texture, timestamp, codec='h264', pixfmt='yuv420p', fps=60, realtime=False,
                          crf=None, preset=None, limit=None):
        import av
        self._touched = system_clock()
        writer = queue = start = None
        width, height = texture.width, texture.height
        has_alpha = texture.components == 4
        config = [width, height, has_alpha, codec, pixfmt, fps, crf, preset, limit]
        if 'video_output' in self._cache:
            writer, queue, start, *cached_config = self._cache['video_output']
            if cached_config != config:
                if queue is not None:
                    logger.debug("Closing output video container as configuration has changed")
                    queue.put(None)
                    writer.join()
                container = queue = writer = start = None
        if start is None:
            self.cleanup()
            if self._path.exists():
                logger.warning("Existing video file will be overwritten: {}", self._path)
            options = {}
            options['color_primaries'] = 'bt709'
            options['color_trc'] = 'bt709'
            options['colorspace'] = 'bt709'
            if crf is not None:
                options['crf'] = str(crf)
            if preset is not None:
                options['preset'] = preset
            try:
                codec = codec.lower()
                if codec not in av.codecs_available:
                    raise ValueError(f"Unrecognised codec '{codec}''")
                av_codec = av.codec.Codec(codec, mode='w')
                if av_codec.type != 'video':
                    raise ValueError(f"'{codec}' not a video codec")
                if av_codec.name == 'hevc_videotoolbox' and has_alpha:
                    options['alpha_quality'] = '1'
                container = av.open(str(self._path), mode='w')
                stream = container.add_stream(av_codec, rate=fps, options=options)
            except Exception as exc:
                logger.opt(exception=exc).error("Unable to open video output: {}", self._path)
                self._cache['video_output'] = False, None, timestamp, *config
                return
            stream.width = width
            stream.height = height
            stream.pix_fmt = pixfmt
            stream.codec_context.time_base = fractions.Fraction(1, fps)
            if av_codec.name == 'libx265' or av_codec.name == 'hevc_videotoolbox':
                stream.codec_context.codec_tag = 'hvc1'
            queue = Queue(maxsize=fps)
            writer = threading.Thread(target=self._video_writer, args=(container, stream, queue))
            writer.start()
            start = timestamp
            self._cache['video_output'] = writer, queue, start, *config
            logger.debug("Beginning {} {} video output{}: {}", av_codec.name, stream.pix_fmt, " (with alpha)" if has_alpha else "", self._path)
        if queue is not None and (not realtime or not queue.full()):
            frame_time = int(round((timestamp - start) * fps))
            if limit is not None and frame_time >= int(round(limit * fps)):
                queue.put(None)
                writer.join()
                self._cache['video_output'] = None, None, start, *config
                return
            frame = av.VideoFrame(width, height, 'rgba' if has_alpha else 'rgb24')
            line_size = frame.planes[0].line_size
            if line_size != width:
                import numpy as np
                components = 4 if has_alpha else 3
                data = np.ndarray((height, width*components), dtype='uint8', buffer=texture.read())
                array = np.empty((height, line_size), dtype='uint8')
                array[:, :width*components] = data
                frame.planes[0].update(array.data)
            else:
                frame.planes[0].update(texture.read())
            frame.pts = frame_time
            try:
                queue.put(frame, block=not realtime)
            except Full:
                pass

    def _video_writer(self, container, stream, queue):
        try:
            while True:
                frame = queue.get()
                for packet in stream.encode(frame):
                    container.mux(packet)
                if frame is None:
                    break
        except Exception as exc:
            logger.opt(exception=exc).error("Error encoding video frame")
        container.close()
        logger.success("Flushed and closed video output file: {}", self._path)

    def __str__(self):
        return str(self._path)


class FileCache:
    def __init__(self):
        self._cache = {}
        self._root = Path('.')

    def clean(self, max_age=DEFAULT_CLEAN_TIME):
        cutoff = system_clock() - max_age
        for path in list(self._cache):
            cache_path = self._cache[path]
            if cache_path._touched < cutoff:
                cache_path.cleanup()
                del self._cache[path]
                logger.trace("Discarded {}", path)

    def set_root(self, path):
        if isinstance(path, CachePath):
            path = path._path
        else:
            path = Path(path)
        if path.exists():
            if not path.is_dir():
                path = path.parent
            self._root = path
        else:
            self._root = Path('.')

    def get_with_root(self, path, root):
        if not isinstance(path, Path):
            path = Path(path)
        if not path.is_absolute():
            if isinstance(root, CachePath):
                root = root._path.parent
            else:
                if not isinstance(root, Path):
                    root = Path(root)
                if root.is_file():
                    root = root.parent
            path = root / path
        absolute = str(path.resolve())
        if (cached := self._cache.get(absolute)) is not None:
            return cached
        self._cache[absolute] = path = CachePath(path, absolute)
        return path

    def __getitem__(self, path):
        return self.get_with_root(path, self._root)


SharedCache = FileCache()




from setuptools import setup, Extension

setup(
    ext_modules=[
        Extension("flitter.language.functions", ["src/flitter/language/functions.pyx"]),
        Extension("flitter.language.noise", ["src/flitter/language/noise.pyx"]),
        Extension("flitter.language.tree", ["src/flitter/language/tree.pyx"]),
        Extension("flitter.language.vm", ["src/flitter/language/vm.pyx"]),
        Extension("flitter.model", ["src/flitter/model.pyx"]),
        Extension("flitter.render.dmx", ["src/flitter/render/dmx.pyx"]),
        Extension("flitter.render.laser", ["src/flitter/render/laser.pyx"]),
        Extension("flitter.render.physics", ["src/flitter/render/physics.pyx"]),
        Extension("flitter.render.window.canvas", ["src/flitter/render/window/canvas.pyx"]),
        Extension("flitter.render.window.canvas3d", ["src/flitter/render/window/canvas3d.pyx"]),
        Extension("flitter.render.window.models", ["src/flitter/render/window/models.pyx"]),
    ]
)



"""
Tests of the flitter controller generic MIDI module
"""

import asyncio
import time
import unittest
from unittest.mock import patch

from loguru import logger

from flitter.clock import system_clock
from flitter.render.controller import midi


logger.disable('flitter.render.controller.midi')


@patch('flitter.render.controller.midi.rtmidi2')
class TestMidiPortCreation(unittest.IsolatedAsyncioTestCase):
    async def test_create_input(self, rtmidi2_mock):
        port = midi.MidiPort('TEST_MIDI_PORT', output=False)
        port._midi_in.open_port.assert_called_with('TEST_MIDI_PORT')

    def test_create_output(self, rtmidi2_mock):
        port = midi.MidiPort('TEST_MIDI_PORT', input=False)
        port._midi_out.open_port.assert_called_with('TEST_MIDI_PORT')

    async def test_create_input_output(self, rtmidi2_mock):
        port = midi.MidiPort('TEST_MIDI_PORT')
        port._midi_in.open_port.assert_called_with('TEST_MIDI_PORT')
        port._midi_out.open_port.assert_called_with('TEST_MIDI_PORT')

    def test_attempt_create_missing(self, rtmidi2_mock):
        rtmidi2_mock.MidiIn.side_effect = RuntimeError()
        rtmidi2_mock.MidiOut.side_effect = RuntimeError()
        with self.assertRaises(ValueError):
            midi.MidiPort('MISSING_PORT')

    def test_attempt_create_missing_input(self, rtmidi2_mock):
        rtmidi2_mock.MidiIn.side_effect = RuntimeError()
        rtmidi2_mock.MidiOut.side_effect = RuntimeError()
        with self.assertRaises(ValueError):
            midi.MidiPort('MISSING_PORT', output=False)

    def test_attempt_create_missing_output(self, rtmidi2_mock):
        rtmidi2_mock.MidiIn.side_effect = RuntimeError()
        rtmidi2_mock.MidiOut.side_effect = RuntimeError()
        with self.assertRaises(ValueError):
            midi.MidiPort('MISSING_PORT', input=False)

    async def test_create_multiple_ports(self, rtmidi2_mock):
        port1 = midi.MidiPort('TEST_MIDI_PORT1')
        port1._midi_in.open_port.assert_called_with('TEST_MIDI_PORT1')
        port1._midi_out.open_port.assert_called_with('TEST_MIDI_PORT1')
        port2 = midi.MidiPort('TEST_MIDI_PORT2')
        port2._midi_in.open_port.assert_called_with('TEST_MIDI_PORT2')
        port2._midi_out.open_port.assert_called_with('TEST_MIDI_PORT2')

    async def test_virtual_port(self, rtmidi2_mock):
        port = midi.MidiPort('TEST_VIRTUAL_PORT', virtual=True)
        port._midi_in.open_virtual_port.assert_called_with('TEST_VIRTUAL_PORT')
        port._midi_out.open_virtual_port.assert_called_with('TEST_VIRTUAL_PORT')


@patch('flitter.render.controller.midi.rtmidi2')
class TestMidiPortClose(unittest.IsolatedAsyncioTestCase):
    def test_close_input_output(self, rtmidi2_mock):
        port = midi.MidiPort('TEST_MIDI_PORT')
        midi_in_mock = port._midi_in
        midi_out_mock = port._midi_out
        port.close()
        midi_in_mock.close_port.assert_called_with()
        midi_out_mock.close_port.assert_called_with()
        self.assertIsNone(port._midi_in)
        self.assertIsNone(port._midi_out)

    def test_close_send_fail(self, rtmidi2_mock):
        port = midi.MidiPort('TEST_MIDI_PORT', input=False)
        port.send_note_off(10)
        port.close()
        with self.assertRaises(TypeError):
            port.send_note_off(10)

    async def test_close_wait_fail(self, rtmidi2_mock):
        port = midi.MidiPort('TEST_MIDI_PORT', output=False)
        await port.wait_event(0.1)
        port.close()
        with self.assertRaises(TypeError):
            await port.wait_event(0.1)


class TestReceiveMidi(unittest.IsolatedAsyncioTestCase):
    def setUp(self):
        self.rtmidi2_patcher = patch('flitter.render.controller.midi.rtmidi2')
        self.rtmidi2_mock = self.rtmidi2_patcher.start()
        self.port = midi.MidiPort('TEST_MIDI_PORT', output=False)

    def tearDown(self):
        self.rtmidi2_patcher.stop()

    def test_send_attempt(self):
        with self.assertRaises(TypeError):
            self.port.send_note_off(10)

    async def test_receive_timeout(self):
        with self.assertRaises(asyncio.TimeoutError):
            await asyncio.wait_for(self.port.wait_event(), 0.1)

    async def test_receive_with_timeout(self):
        event = await self.port.wait_event(0.1)
        self.assertIsNone(event)

    async def test_receive_note_off(self):
        start = system_clock()
        self.port._midi_in.callback(bytes([0x80, 0x0a]), 0.0)
        end = system_clock()
        event = await self.port.wait_event()
        self.assertIsInstance(event, midi.NoteOffEvent)
        self.assertEqual(event.channel, 0)
        self.assertTrue(start < event.timestamp < end)
        self.assertEqual(event.note, 10)

    async def test_receive_note_on(self):
        start = system_clock()
        self.port._midi_in.callback(bytes([0x91, 0x0a, 0x7f]), 0.0)
        end = system_clock()
        event = await self.port.wait_event()
        self.assertIsInstance(event, midi.NoteOnEvent)
        self.assertEqual(event.channel, 1)
        self.assertTrue(start < event.timestamp < end)
        self.assertEqual(event.note, 10)
        self.assertEqual(event.velocity, 127)

    async def test_receive_note_pressure(self):
        start = system_clock()
        self.port._midi_in.callback(bytes([0xa2, 0x0a, 0x37]), 0.0)
        end = system_clock()
        event = await self.port.wait_event()
        self.assertIsInstance(event, midi.NotePressureEvent)
        self.assertEqual(event.channel, 2)
        self.assertTrue(start < event.timestamp < end)
        self.assertEqual(event.note, 10)
        self.assertEqual(event.pressure, 55)

    async def test_receive_control_change(self):
        start = system_clock()
        self.port._midi_in.callback(bytes([0xb3, 0x10, 0x21]), 0.0)
        end = system_clock()
        event = await self.port.wait_event()
        self.assertIsInstance(event, midi.ControlChangeEvent)
        self.assertEqual(event.channel, 3)
        self.assertTrue(start < event.timestamp < end)
        self.assertEqual(event.control, 16)
        self.assertEqual(event.value, 33)

    async def test_receive_channel_pressure(self):
        start = system_clock()
        self.port._midi_in.callback(bytes([0xd4, 0x37]), 0.0)
        end = system_clock()
        event = await self.port.wait_event()
        self.assertIsInstance(event, midi.ChannelPressureEvent)
        self.assertEqual(event.channel, 4)
        self.assertTrue(start < event.timestamp < end)
        self.assertEqual(event.pressure, 55)

    async def test_receive_pitch_bend(self):
        start = system_clock()
        self.port._midi_in.callback(bytes([0xe5, 0x72, 0x7f]), 0.0)
        end = system_clock()
        event = await self.port.wait_event()
        self.assertIsInstance(event, midi.PitchBendEvent)
        self.assertEqual(event.channel, 5)
        self.assertTrue(start < event.timestamp < end)
        self.assertEqual(event.value, 16370)

    async def test_receive_timing(self):
        start = system_clock()
        self.port._midi_in.callback(bytes([0x80, 0x0a]), 0.0)
        time.sleep(1/20)
        self.port._midi_in.callback(bytes([0x80, 0x0a]), 1/32)
        event1 = await self.port.wait_event()
        event2 = await self.port.wait_event()
        end = system_clock()
        self.assertTrue(start < event1.timestamp < end)
        self.assertEqual(event2.timestamp, event1.timestamp + 1/32)


class TestSendMidi(unittest.IsolatedAsyncioTestCase):
    def setUp(self):
        self.rtmidi2_patcher = patch('flitter.render.controller.midi.rtmidi2')
        self.rtmidi2_mock = self.rtmidi2_patcher.start()
        self.port = midi.MidiPort('TEST_MIDI_PORT', input=False, sysex_id=(0x01, 0x02, 0x03))

    def tearDown(self):
        self.rtmidi2_patcher.stop()

    async def test_receive_attempt(self):
        with self.assertRaises(TypeError):
            await self.port.wait_event()

    def test_send_note_off(self):
        self.port.send_note_off(10)
        self.port._midi_out.send_raw.assert_called_with(0x80, 0x0a)

    def test_send_note_on(self):
        self.port.send_note_on(10, channel=1)
        self.port._midi_out.send_raw.assert_called_with(0x91, 0x0a, 0x7f)
        self.port.send_note_on(10, 55, channel=2)
        self.port._midi_out.send_raw.assert_called_with(0x92, 0x0a, 0x37)

    def test_send_note_pressure(self):
        self.port.send_note_pressure(10, 55, channel=2)
        self.port._midi_out.send_raw.assert_called_with(0xa2, 0x0a, 0x37)

    def test_send_control_change(self):
        self.port.send_control_change(16, 33, channel=3)
        self.port._midi_out.send_raw.assert_called_with(0xb3, 0x10, 0x21)

    def test_send_program_change(self):
        self.port.send_program_change(19, channel=9)
        self.port._midi_out.send_raw.assert_called_with(0xc9, 0x13)

    def test_send_channel_pressure(self):
        self.port.send_channel_pressure(55, channel=4)
        self.port._midi_out.send_raw.assert_called_with(0xd4, 0x37)

    def test_send_pitch_bend_change(self):
        self.port.send_pitch_bend_change(16370, channel=5)
        self.port._midi_out.send_raw.assert_called_with(0xe5, 0x72, 0x7f)

    def test_send_sysex(self):
        self.port.send_sysex(0x04, 0x05, 0x06)
        self.port._midi_out.send_raw.assert_called_with(0xf0, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0xf7)

    def test_send_sysex_explicit_id(self):
        self.port.send_sysex(0x04, 0x05, 0x06, sysex_id=(0x71, 0x72, 0x73))
        self.port._midi_out.send_raw.assert_called_with(0xf0, 0x71, 0x72, 0x73, 0x04, 0x05, 0x06, 0xf7)

    def test_send_clock(self):
        self.port.send_clock()
        self.port._midi_out.send_raw.assert_called_with(0xf8)

    def test_send_start(self):
        self.port.send_start()
        self.port._midi_out.send_raw.assert_called_with(0xfa)

    def test_send_continue(self):
        self.port.send_continue()
        self.port._midi_out.send_raw.assert_called_with(0xfb)

    def test_send_stop(self):
        self.port.send_stop()
        self.port._midi_out.send_raw.assert_called_with(0xfc)

    def test_send_active(self):
        self.port.send_active()
        self.port._midi_out.send_raw.assert_called_with(0xfe)

    def test_send_reset(self):
        self.port.send_reset()
        self.port._midi_out.send_raw.assert_called_with(0xff)



"""
Push LED color management
"""

import colorsys


class Palette:
    def rgb_to_index(self, r, g, b):
        """Translate RGB values in range 0..1 to an index in range 0..127"""
        raise NotImplementedError()

    def index_to_rgb_led(self, i):
        """Translate RGB index in range 0..127 to LED RGB values in range 0..255"""
        raise NotImplementedError()

    def white_to_index(self, w):
        """Translate white value in range 0..1 to an index in range 0..127"""
        raise NotImplementedError()

    def index_to_white_led(self, i):
        """Translate white index in range 0..127 to LED white value in range 0..255"""
        raise NotImplementedError()


class SimplePalette(Palette):
    """
    A palette that provides 128 levels of white for white LEDs and 5x5x5 colors
    for RGB LEDs, with a 2.2 gamma applied.

    The white LED transfer function approximates 2.2 gamma, but with some
    linearity to reduce stepping at the dark end.
    """

    def rgb_to_index(self, r, g, b):
        r, g, b = (int(round(min(max(0, c), 0.8) * 5)) for c in (r, g, b))
        return ((r * 5) + g) * 5 + b

    def index_to_rgb_led(self, i):
        i = min(max(0, int(i)), 124)
        r, g, b = i // 25, (i // 5) % 5, i % 5
        return tuple(int(((c / 4) ** 2.2) * 255) for c in (r, g, b))

    def white_to_index(self, w):
        return int(min(max(0, w), 0.9921875) * 128)

    def index_to_white_led(self, i):
        value = min(max(0, int(i)), 127)
        return value // 4 + int(224 * (value / 127) ** 2.4)


class PrimaryPalette(SimplePalette):
    """
    This palette changes the simple palette for RGB LEDs to provide 32 levels
    of white and 17 levels each of red, green, blue, yellow, cyan and magenta
    (including a shared value for black). This is a good palette if you don't
    plan to use a lot of colours, but do want to do pulsing effects with the
    LED brightness.

    Again, the LED transfer function approximates 2.2 gamma, but with some
    linearity to reduce stepping at the dark end.
    """

    def rgb_to_index(self, r, g, b):
        r, g, b = (min(max(0, c), 1) for c in (r, g, b))
        value = max(r, g, b)
        if value == 0:
            return 0
        r, g, b = (int(round(c / value)) for c in (r, g, b))
        if r == g == b:
            value = int(min(value * 32, 31))
            return value if value < 16 else value + 0b1100000
        value = int(min(value * 17, 16))
        if value == 0:
            return 0
        return (r << 4) + (g << 5) + (b << 6) + (value - 1)

    def index_to_rgb_led(self, i):
        i = min(max(0, int(i)), 127)
        rgb = i >> 4
        if rgb in (0b000, 0b111):
            value = i & 0b11111
            value += int(224 * (value / 31)**2.4)
            return (value, value, value)
        value = (i & 0b1111) + 1
        value += int(239 * (value / 16)**2.4)
        return (value if rgb & 0b001 else 0, value if rgb & 0b010 else 0, value if rgb & 0b100 else 0)


class HuePalette(SimplePalette):
    """
    This palette provides 32 shades of white, but then 24 different hues at
    2 levels of saturation (0.5 and 1) and 2 levels of brightness (0.5 and 1).
    This is a good palette if you're content with just a couple of levels of
    brightness - e.g., to show toggled/touched status - but want lots of
    different colours.
    """

    def rgb_to_index(self, r, g, b):
        h, s, v = colorsys.rgb_to_hsv(r, g, b)
        s = int(round(s * 2))
        if s == 0:
            return int(min(max(0, v), 0.96875) * 32)
        h = int(round(h * 24)) % 24
        v = int(round(v * 2))
        if v == 0:
            return 0
        return ((v - 1) * 48) + ((s - 1) * 24) + h + 32

    def index_to_rgb_led(self, i):
        i = min(max(0, int(i)), 127)
        if i < 32:
            v = i + int(224 * (i / 31)**2.4)
            return (v, v, v)
        i -= 32
        h = i % 24 / 24
        s = ((i // 24) % 2 + 1) / 2
        v = ((i // 48) + 1) / 2
        return tuple(int(c**2.2 * 255) for c in colorsys.hsv_to_rgb(h, s, v))
